import re
import functools
import itertools

import pysam
import pyranges as pr
import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
import Bio.Seq

import handygenome.signature.misc as libsigmisc
import handygenome.deco as deco
import handygenome.refgenome.refgenome as refgenome


"""Issues with pyranges.PyRanges.intersect:
    - "Start" and "End" columns of pyranges.PyRanges object are by 
        default "int32".
    - "intersect" and "join" methods results in error if "Start" and "End"
        columns of "self" gr and "other" gr are of different types.
    - When loading breakends data with missing values, 
        - pd.DataFrame is created with POS1/POS2 columns with nan.
        - After removing nan, the columns are coerced into integer type.
        - At that time, dtype must be "int32".
"""

SV_COLORMAP = {
    'DEL': 'red',
    'DUP': 'green',
    'INV': 'aquamarine',
    'TRA': 'gold',
    np.nan: 'black',
}

CYTOBAND_COLORMAP = {
    "gneg": "#FFFFFF00", 
    "gpos25": "#EEEEEE", 
    "gpos50": "#BBBBBB", 
    "gpos75": "#777777", 
    "gpos100": "#000000", 
    "gvar": "#FFFFFF00", 
    "stalk": "#C01E27", 
    "acen": "#D82322",
}

CYTOBAND_PATHS = refgenome.RefverDict({
    'GRCh37': '/home/users/pjh/References/cytoband/cytoBand_hg19.bed',
    'GRCh38': '/home/users/pjh/References/cytoband/cytoBand_hg38.bed',
})


def read_cytoband(cytoband_path, refver):
    result = pd.read_table(
        cytoband_path, 
        header=None,
        names=['Chromosome', 'Start', 'End', 'Name', 'Stain']
    )
    result['Color'] = result['Stain'].apply(lambda x: CYTOBAND_COLORMAP[x])

    refver = refgenome.RefverDict.standardize(refver)
    if refver == 'GRCh37':
        result['Chromosome'] = result['Chromosome'].apply(
            lambda x: re.sub('^chr', '', x)
        )

    return result


def check_column_existence(gr, names):
    for key in names:
        if key not in gr.columns:
            raise Exception(f'"{key}" must be included in columns')


def linspace_rounded(start, stop, num):
    width = (stop - start) / num
    decimals = int(-np.log10(width))
    return np.linspace(start, stop, num).round(decimals)


class Sector:
    def __init__(self, chrom, start0, end0):
        self.chrom = chrom
        self.start0 = start0
        self.end0 = end0

    def __repr__(self):
        return f'<Sector object(chrom={self.chrom}, start0={self.start0}, end0={self.end0}>'

    @property
    def id(self):
        return (self.chrom, self.start0, self.end0)

    @property
    def length(self):
        return self.end0 - self.start0

    def get_label(self, short_for_full_chromosome=True, chromdict=None):
        if short_for_full_chromosome and (chromdict is None):
            raise Exception(f'If "short_for_full_chromosome" argument is True, "chromdict" argument must be set.')

        if short_for_full_chromosome:
            if self.start0 == 0 and self.end0 == chromdict[self.chrom]:
                return self.chrom
            else:
                return f'{self.chrom}_{self.start0}_{self.end0}'
        else:
            return f'{self.chrom}_{self.start0}_{self.end0}'


class SectorList(list):
    def __init__(self, chroms=list(), start0s=list(), end0s=list()):
        for chrom, start0, end0 in zip(chroms, start0s, end0s):
            self.append(Sector(chrom, start0, end0))

    def add_sector(self, chrom, start0, end0):
        self.append(Sector(chrom, start0, end0))

    @classmethod
    def from_chromdict(cls, chromdict, assembled_only=True):
        chroms = list()
        start0s = list()
        end0s = list()
        for chrom, length in chromdict.items():
            if assembled_only:
                if not refgenome.PAT_ASSEMBLED_CHROM.fullmatch(chrom):
                    continue

            chroms.append(chrom)
            start0s.append(0)
            end0s.append(length)

        return cls(chroms, start0s, end0s)

    @classmethod
    def from_refver(cls, refver, assembled_only=True):
        chromdict = refgenome.ChromDict.from_refver(refver)
        return cls.from_chromdict(chromdict, assembled_only=assembled_only)

    @classmethod
    def from_gr(cls, gr):
        chroms = list()
        start0s = list()
        end0s = list()
        for idx, row in gr.df.iterrows():
            chroms.append(row['Chromosome'])
            start0s.append(row['Start'])
            end0s.append(row['End'])
        return cls(chroms, start0s, end0s)

    @functools.cached_property
    def dict(self):
        return {x.id: x for x in self}

    @functools.cached_property
    def gr(self):
        chroms = list()
        starts = list()
        ends = list()
        ids = list()
        
        for x in self:
            chroms.append(x.chrom)
            starts.append(x.start0)
            ends.append(x.end0)
            ids.append(x.id)

        result = pr.from_dict(
            {
                'Chromosome': chroms,
                'Start': starts,
                'End': ends,
                'ID': ids,
            },
            int64=False,
        )

        return result

    def get_ids(self):
        return [x.id for x in self]


class Circos:
    @deco.get_deco_num_set(('refver', 'chromdict'), 1)
    def __init__(self, refver=None, chromdict=None, sectorlist=None, ylim=(0, 1), sector_interspace_deg=2, figsize=(10, 10), angle_range_deg=(0, 360), theta_offset_deg=90, theta_direction=-1):
        """Args:
            sector_interspace_deg: A scalar or a sequence type with the same 
                length as sectorlist. Each element represents the space BEFORE
                to the sector of the same index.
        """
        # chromdict
        if refver is not None:
            self.chromdict = refgenome.ChromDict.from_refver(refver)
        elif chromdict is not None:
            self.chromdict = chromdict
        # sectorlist
        if sectorlist is None:
            self.sectorlist = SectorList.from_chromdict(self.chromdict, assembled_only=True)
        else:
            self.sectorlist = sectorlist
        # sector interspace
        self.sector_interspace_deg = self.arghandler_sector_interspace_deg(sector_interspace_deg)
        self.sector_interspace_rad = np.deg2rad(sector_interspace_deg)
        # others 
        self.ylim = ylim
        self.figsize = figsize
        self.angle_range_deg = self.get_angle_range(angle_range_deg)
        self.angle_length_deg = self.angle_range_deg[1] - self.angle_range_deg[0]
        self.theta_offset_deg = theta_offset_deg
        self.theta_direction = theta_direction
        # postprocess
        #self.init_figure()
        self.set_sector_angles()

    @staticmethod
    def get_angle_range(raw_angle_range_deg):
        if raw_angle_range_deg[0] == raw_angle_range_deg[1]:
            raise Exception(f'Start and end of "angle_range" must be different.')

        start_angle = np.mod(raw_angle_range_deg[0], 360)
        end_angle = np.mod(raw_angle_range_deg[1], 360)
        if start_angle == end_angle:
            return (start_angle, end_angle + 360)
        elif end_angle > start_angle:
            return (start_angle, end_angle)
        elif end_angle < start_angle:
            return (start_angle, end_angle + 360)

    def set_sector_angles(self):
        sector_interspace_deg_sum = self.sector_interspace_deg.sum()
        if sector_interspace_deg_sum > self.angle_length_deg:
            raise Exception(f'Sum of interspaces({sector_interspace_deg_sum}) is greater than the total circos angle({self.angle_length_deg}). Choose a smaller "sector_interspace_deg".')

        sector_length_angle_deg_sum = self.angle_length_deg - sector_interspace_deg_sum
        sector_length_sum = sum(x.length for x in self.sectorlist)

        cursor_deg = self.angle_range_deg[0]
        sector_angles_rad = dict()
        #sector_angles_deg = dict()
        for sector, interspace_deg in zip(self.sectorlist, self.sector_interspace_deg):
            # now cursor_deg indicates "sector start position" - "interspace"
            sector_start_angle_deg = cursor_deg + interspace_deg
            sector_end_angle_deg = sector_start_angle_deg + sector_length_angle_deg_sum * (sector.length / sector_length_sum)

            sector_angle_pair_deg = (sector_start_angle_deg, sector_end_angle_deg)
            sector_angle_pair_rad = tuple(np.deg2rad(sector_angle_pair_deg))
            #sector_angles_deg[sector.id] = sector_angle_pair_deg
            sector_angles_rad[sector.id] = sector_angle_pair_rad

            cursor_deg = sector_end_angle_deg

        self.sector_angles_rad = sector_angles_rad
        #self.sector_angles_deg = sector_angles_deg

    # drawing methods
#    def init_figure(self):  # deprecated
#        self.figure = plt.figure(figsize=self.figsize)
#        self.ax = self.figure.add_axes([0, 0, 1, 1], polar=True)
#
#        self.ax.set_theta_offset(np.deg2rad(self.theta_offset_deg))
#        self.ax.set_theta_direction(self.theta_direction)
#        self.ax.set_xlim(*np.deg2rad(self.angle_range_deg))
#        self.ax.set_ylim(self.ylim[0], self.ylim[1])
#        self.ax.spines['polar'].set_visible(False)
#        self.ax.xaxis.set_ticks([])
#        self.ax.xaxis.set_ticklabels([])
#        self.ax.yaxis.set_ticks([])
#        self.ax.yaxis.set_ticklabels([])
#
#        plt.close()

    def init_plot(self, ax):
        ax.set_theta_offset(np.deg2rad(self.theta_offset_deg))
        ax.set_theta_direction(self.theta_direction)
        ax.set_xlim(*np.deg2rad(self.angle_range_deg))
        ax.set_ylim(self.ylim[0], self.ylim[1])
        ax.spines['polar'].set_visible(False)
        ax.xaxis.set_ticks([])
        ax.xaxis.set_ticklabels([])
        ax.yaxis.set_ticks([])
        ax.yaxis.set_ticklabels([])

        plt.close()

    def draw_sector_titles(self, ax, short_for_full_chromosome=True, radial_pos=1.15, fontsize=20, perpendicular=False):
        for sector in self.sectorlist: 
            label = sector.get_label(short_for_full_chromosome=short_for_full_chromosome, chromdict=self.chromdict)
            sector_angle_pair_rad = self.sector_angles_rad[sector.id]
            position_rad = (sector_angle_pair_rad[0] + sector_angle_pair_rad[1]) / 2
            position_deg = self.to_canonical_degrees(position_rad)
            if perpendicular:
                rot = self.get_rotation_perpendicular(position_deg)
            else:
                rot = self.get_rotation_upright(position_deg)
            # draw text
            ax.text(position_rad, radial_pos, label, rotation=rot, ha='center', va='center', fontsize=fontsize)

    def draw_poslabels_within_sector(self, ax, interspace_deg=30, radial_pos=1.10, fontsize=10):
        tick_angles_deg = np.arange(start=self.angle_range_deg[0], stop=self.angle_range_deg[1], step=interspace_deg)
        tick_angles_rad = np.deg2rad(tick_angles_deg)
        for angle_rad in tick_angles_rad:
            data_position = self.get_data_position(angle_rad)
            if data_position is None:
                continue
            label = f'{data_position:,}'
            canonical_degree = self.to_canonical_degrees(angle_rad)
            rot = self.get_rotation_perpendicular(canonical_degree)
            ax.text(angle_rad, radial_pos, label, rotation=rot, ha='center', va='center', fontsize=fontsize)

    def draw_cytoband(self, ax, refver, radial_range=(0.94, 1)):
        cytoband_df = read_cytoband(CYTOBAND_PATHS[refver], refver)
        self.draw_frame(ax, radial_range)
        self.draw_bar(ax, cytoband_df, radial_range)

    def draw_point_mutation(self, ax, data_df, radial_range=(0.8, 0.94), size=5, draw_gridlabel=True):
        def make_colors(df):
            colormap = libsigmisc.COLORS_SBS6
            colorlist = list()
            for idx, row in df.iterrows():
                ref = row['REF']
                alt = row['ALT']
                if len(ref) == len(alt) and len(ref) == 1:
                    if ref in 'CT':
                        color = colormap[ref + '>' + alt]
                    elif ref in 'AG':
                        ref = Bio.Seq.complement(ref)
                        alt = Bio.Seq.complement(alt)
                        color = colormap[ref + '>' + alt]
                    else:
                        color = colormap['other']
                else:
                    color = colormap['other']

                colorlist.append(color)

            return colorlist

        # main
        where_to_draw_label = self.arghandler_draw_gridlabel(draw_gridlabel)
        check_column_existence(data_df, ('CHROM', 'POS', 'REF', 'ALT', 'VAF'))

        data_df['Chromosome'] = data_df['CHROM']
        data_df['Pos0'] = data_df['POS'] - 1
        data_df['Y'] = data_df['VAF']
        data_df['Color'] = make_colors(data_df)

        ymin = 0
        ymax = 1
        gridline_ys = [0.25, 0.5, 0.75]
        gridlabel_ys = [0, 0.5, 1] if where_to_draw_label.any() else None
        self.draw_frame(ax, radial_range, gridline_ys=gridline_ys, gridlabel_ys=gridlabel_ys, ymin=ymin, ymax=ymax, where_to_draw_label=where_to_draw_label)
        self.draw_scatter(ax, data_df, radial_range, ymin=ymin, ymax=ymax, size=size)

    def draw_copynumber(self, ax, data_df, radial_range, color_A='red', color_B='blue', ymin=None, ymax=None, ymax_quantile=0.95, y_offset=0.1, linewidth=0.5, draw_gridlabel=True):
        where_to_draw_label = self.arghandler_draw_gridlabel(draw_gridlabel)
        check_column_existence(data_df, ('Chromosome', 'Start', 'End', 'A', 'B'))
        # filter out NaN rows
        data_df = data_df.loc[np.logical_not(data_df['A'].isnull()), :]
        data_df = data_df.loc[np.logical_not(data_df['B'].isnull()), :]
        # A
        data_df_A = data_df.copy()
        data_df_A['Y'] = data_df_A['A']
        zero_rows = data_df_A['Y'] == 0
        data_df_A.loc[zero_rows, 'Y'] += 2 * y_offset
        data_df_A.loc[np.logical_not(zero_rows), 'Y'] += y_offset
        data_df_A['Color'] = color_A
        # B
        data_df_B = data_df.copy()
        data_df_B['Y'] = data_df_B['B']
        zero_rows = data_df_A['Y'] == 0
        data_df_B.loc[zero_rows, 'Y'] += y_offset
        data_df_B.loc[np.logical_not(zero_rows), 'Y'] -= y_offset
        data_df_B['Color'] = color_B
        # set limits
        if ymin is None:
            ymin = 0
        if ymax is None:
            data_df_modified = self.modify_data(data_df, make_radius=False)
            all_ys = pd.Series(itertools.chain(data_df_modified['A'], data_df_modified['B']))
            all_ys = all_ys.loc[np.logical_not(all_ys.isnull())]
            ymax = np.quantile(all_ys, ymax_quantile)
            ymax = max(ymax, 6)
        # draw frame
        gridline_ys = np.arange(start=np.rint(ymin), stop=np.rint(ymax), step=1, dtype='int')
        if where_to_draw_label.any():
            gridlabel_ys = np.arange(start=np.rint(ymin), stop=np.rint(ymax), step=2, dtype='int')
        else:
            gridlabel_ys = None

        self.draw_frame(ax, radial_range, gridline_ys=gridline_ys, gridlabel_ys=gridlabel_ys, ymin=ymin, ymax=ymax, where_to_draw_label=where_to_draw_label)
        # draw arcs
        self.draw_arc(ax, data_df_A, radial_range, linewidth=linewidth, ymin=ymin, ymax=ymax)
        self.draw_arc(ax, data_df_B, radial_range, linewidth=linewidth, ymin=ymin, ymax=ymax)

    def draw_depth_scatter(self, ax, data_df, radial_range, color='black', ymin=None, ymax=None, ymax_quantile=0.99, size=1, gridline_num=8, draw_gridlabel=True):
        where_to_draw_label = self.arghandler_draw_gridlabel(draw_gridlabel)
        check_column_existence(data_df, ('Chromosome', 'Start', 'End', 'Depth'))

        data_df['Pos0'] = ((data_df['Start'] + data_df['End']) * 0.5).apply(int)
        data_df['Y'] = data_df['Depth']
        data_df['Color'] = color
        # set limits
        if ymin is None:
            ymin = 0
        if ymax is None:
            data_df_modified = self.modify_data(data_df, make_radius=False)
            ymax = np.quantile(data_df_modified['Depth'], ymax_quantile)
            ymax = max(ymax, 60)
        # draw frame
        gridline_start = np.floor(ymin)
        gridline_stop = np.ceil(ymax)
        gridline_ys = np.rint(np.linspace(start=gridline_start, stop=gridline_stop, num=gridline_num, endpoint=True)).astype('int')
        if where_to_draw_label.any():
            gridlabel_ys = gridline_ys[::2]
        else:
            gridlabel_ys = None

        self.draw_frame(ax, radial_range, gridline_ys=gridline_ys, gridlabel_ys=gridlabel_ys, ymin=ymin, ymax=ymax, where_to_draw_label=where_to_draw_label)
        # draw dots
        self.draw_scatter(ax, data_df, radial_range, ymin=ymin, ymax=ymax, size=size)

    draw_depth = draw_depth_scatter

    def draw_breakends(self, ax, data_df, radial_pos, linewidth=0.5):
        check_column_existence(data_df, ('CHROM_site1', 'POS0_site1', 'CHROM_site2', 'POS0_site2', 'Color'))
        self.draw_chord(ax, data_df, radial_pos, linewidth=linewidth)

    #################

    @staticmethod
    def get_radius_list(y, radial_range, ymin=None, ymax=None):
        y = np.array(y)

        if ymin is None:
            ymin = min(y)
        if ymax is None:
            ymax = max(y)

        y_range = ymax - ymin
        radial_length = radial_range[1] - radial_range[0]
        bottom = radial_range[0]
        dataspace_portion = (y - ymin) / y_range
        return bottom + radial_length * dataspace_portion

    @staticmethod
    def get_angle(data_position, sector, sector_start_angle_rad, sector_length_angle_rad):
        portion_within_sector = (data_position - sector.start0) / sector.length
        angle_within_sector_rad = sector_length_angle_rad * portion_within_sector
        return sector_start_angle_rad + angle_within_sector_rad
    
    def get_data_position(self, angle_rad):
        # find a corresponding sector id
        found_a_sector = False
        for sector_id, angle_pair_rad in self.sector_angles_rad.items():
            if angle_rad >= angle_pair_rad[0] and angle_rad < angle_pair_rad[1]:
                found_a_sector = True
                break
        if not found_a_sector:
            return None

        # calculate data position
        angle_within_sector_rad = angle_rad - angle_pair_rad[0]
        sector_length_angle_rad = angle_pair_rad[1] - angle_pair_rad[0]
        portion_within_sector = angle_within_sector_rad / sector_length_angle_rad
        sector = self.sectorlist.dict[sector_id]
        data_position = int(sector.start0 + portion_within_sector * sector.length)
        return data_position

    def to_canonical_degrees(self, angle_rad):
        return self.theta_offset_deg + np.rad2deg(angle_rad) * self.theta_direction

    @staticmethod
    def get_rotation_upright(canonical_degree):
        rot = canonical_degree - 90
        canonical_degree_mod = np.mod(canonical_degree, 360)
        if canonical_degree_mod > 180 and canonical_degree_mod < 360:
            rot += 180
        return rot

    @staticmethod
    def get_rotation_perpendicular(canonical_degree):
        rot = canonical_degree
        canonical_degree_mod = np.mod(canonical_degree, 360)
        if canonical_degree_mod > 90 and canonical_degree_mod < 270:
            rot += 180
        return rot

    def modify_data(self, data_df, make_radius=False, radial_range=None, ymin=None, ymax=None):
        def sanity_check(data_df, make_radius, radial_range):
            check_column_existence(data_df, ('Chromosome', 'Start', 'End'))
            if make_radius:
                if radial_range is None:
                    raise Exception(f'If "make_radius" is True, "radial_range" must be given.')
                if 'Y' not in data_df.columns:
                    raise Exception(f'If "make_radius" is True, "Y" column must be included in "data_df".')

        # sanity check
        sanity_check(data_df, make_radius, radial_range)
        # save original data_df columns
        columns = data_df.columns
        # perform join
        data_gr = pr.PyRanges(df=data_df, int64=False)
        gr_joined = data_gr.join(self.sectorlist.gr).new_position('intersection')
        if gr_joined.empty:
            result_columns = list(columns)
            result_columns.extend(['Start_rad', 'End_rad'])
            if make_radius:
                result_columns.append('Radius')
            result = pd.DataFrame(columns=result_columns)
        else:
            data_start_angles_rad = list()
            data_end_angles_rad = list()
            for idx, row in gr_joined.df.iterrows():
                sector_id = row['ID']
                sector = self.sectorlist.dict[sector_id]
                sector_angle_pair_rad = self.sector_angles_rad[sector_id]
                sector_start_angle_rad = sector_angle_pair_rad[0]
                sector_length_angle_rad = sector_angle_pair_rad[1] - sector_angle_pair_rad[0]

                data_start_angles_rad.append(
                    self.get_angle(row['Start'], sector, sector_start_angle_rad, sector_length_angle_rad)
                )
                data_end_angles_rad.append(
                    self.get_angle(row['End'], sector, sector_start_angle_rad, sector_length_angle_rad)
                )

            result = gr_joined.df
            result.insert(result.shape[1], 'Start_rad', data_start_angles_rad)
            result.insert(result.shape[1], 'End_rad', data_end_angles_rad)

            if make_radius:
                radius_list = self.get_radius_list(result['Y'], radial_range, ymin=ymin, ymax=ymax)
                result.insert(result.shape[1], 'Radius', radius_list)
                # filter out data out of radial_range
                result = result.loc[
                    (result['Radius'] <= radial_range[1]) & (result['Radius'] >= radial_range[0]),
                    :
                ]

        return result

    def modify_breakends_data(self, data_df):

        def set_site1_coords(data):
            if isinstance(data, pr.PyRanges):
                # without coercing into int32, "intersect" method results in error
                data = data.assign('Chromosome', lambda df: df['CHROM_site1'])
                data = data.assign('Start', lambda df: df['POS0_site1'].astype('int32'))
                data = data.assign('End', lambda df: (df['POS0_site1'] + 1).astype('int32'))
            elif isinstance(data, pd.DataFrame):
                data['Chromosome'] = data['CHROM_site1']
                data['Start'] = data['POS0_site1']
                data['End'] = data['POS0_site1'] + 1
            return data

        def set_site2_coords(data):
            if isinstance(data, pr.PyRanges):
                # without coercing into int32, "intersect" method results in error
                data = data.assign('Chromosome', lambda df: df['CHROM_site2'])
                data = data.assign('Start', lambda df: df['POS0_site2'].astype('int32'))
                data = data.assign('End', lambda df: (df['POS0_site2'] + 1).astype('int32'))
            elif isinstance(data, pd.DataFrame):
                data['Chromosome'] = data['CHROM_site2']
                data['Start'] = data['POS0_site2']
                data['End'] = data['POS0_site2'] + 1
            return data

        # sanity check
        check_column_existence(data_df, ('CHROM_site1', 'POS0_site1', 'CHROM_site2', 'POS0_site2', 'Color'))
        # remove nan rows - those whose bnd2 is unknown (e.g. telomere)
        data_df = data_df.loc[np.logical_not(data_df['POS0_site1'].isna()), :]
        data_df = data_df.loc[np.logical_not(data_df['POS0_site2'].isna()), :]
        # prepare data_df for turning into a pyranges object
        data_df['Chromosome'] = '1'
        data_df['Start'] = 0
        data_df['End'] = 1
        # initialize gr
        data_gr = pr.PyRanges(df=data_df, int64=False)
        # intersect with site1
        data_gr = set_site1_coords(data_gr)
        data_gr = data_gr.intersect(self.sectorlist.gr)
        if data_gr.empty:
            return pd.DataFrame()
        # intersect with site2
        data_gr = set_site2_coords(data_gr)
        data_gr = data_gr.intersect(self.sectorlist.gr)
        if data_gr.empty:
            return pd.DataFrame()
        # run modify_data with site2
        data_df = self.modify_data(data_gr.df, make_radius=False)
        data_df.insert(data_df.shape[1], 'Angle_site2_rad', data_df['Start_rad'])
        data_df = data_df.drop(columns=['Start_rad', 'End_rad', 'ID', 'Start_b', 'End_b'])
        # run modify_data with site1
        data_df = set_site1_coords(data_df)
        data_df = self.modify_data(data_df, make_radius=False)
        data_df.insert(data_df.shape[1], 'Angle_site1_rad', data_df['Start_rad'])
        data_df = data_df.drop(columns=['Start_rad', 'End_rad', 'ID', 'Start_b', 'End_b'])
        # finally drop unused columns
        data_df = data_df.drop(columns=['Chromosome', 'Start', 'End'])

        return data_df

    def draw_frame(self, ax, radial_range, gridline_ys=None, gridlabel_ys=None, gridlabels=None, gridlabel_size=None, ymin=None, ymax=None, linewidth_grid=0.5, linewidth_spine=1, color_spine='black', color_grid='black', where_to_draw_label=None):
        # sanity checks
        if (gridlabel_ys is not None) and (gridlabels is not None):
            if len(gridlabel_ys) != len(gridlabels):
                raise Exception(f'Lengths of "gridlabel_ys" and "gridlabels" must be equal.')

        # determine if grids will be drawn
        draw_grids = (gridline_ys is not None) or (gridlabel_ys is not None)

        # set default gridlabels
        if (gridlabel_ys is not None) and (gridlabels is None):
            gridlabels = [str(x) for x in gridlabel_ys]

        # set default ymin and ymax
        if draw_grids:
            all_ys = list(
                itertools.chain(
                    (list() if gridline_ys is None else gridline_ys),
                    (list() if gridlabel_ys is None else gridlabel_ys)
                )
            )
            if ymin is None:
                ymin = min(all_ys)
            if ymax is None:
                ymax = max(all_ys)

        # set default "where_to_draw_label"
        if where_to_draw_label is None:
            where_to_draw_label = np.repeat(True, len(self.sectorlist))

        # set other parameters
        spine_height = radial_range[1] - radial_range[0]
        spine_bottom = radial_range[0]

        if gridline_ys is None:
            #gridline_radii = None
            gridline_diameters = None
        else:
            #gridline_radii = self.get_radius_list(gridline_ys, radial_range, ymin=ymin, ymax=ymax)
            gridline_diameters = 2 * self.get_radius_list(gridline_ys, radial_range, ymin=ymin, ymax=ymax)

        if gridlabel_ys is None:
            gridlabel_radii = None
            #gridlabel_diameters = None
        else:
            gridlabel_radii = self.get_radius_list(gridlabel_ys, radial_range, ymin=ymin, ymax=ymax)
            #gridlabel_diameters = 2 * gridlabel_radii

        # draw
        for idx, (sector, draw_label) in enumerate(zip(self.sectorlist, where_to_draw_label)):
            # draw spines
            sector_start_angle_rad, sector_end_angle_rad = self.sector_angles_rad[sector.id]
            width = sector_end_angle_rad - sector_start_angle_rad
            ax.bar(x=sector_start_angle_rad, height=spine_height, width=width, bottom=spine_bottom, align='edge', fill=False, edgecolor=color_spine, linewidth=linewidth_spine)

            if draw_grids:
                # set thetas
                if self.theta_direction == 1:
                    theta1 = self.to_canonical_degrees(sector_start_angle_rad)
                    theta2 = self.to_canonical_degrees(sector_end_angle_rad)
                    start_theta = theta1
                elif self.theta_direction == -1:
                    theta1 = self.to_canonical_degrees(sector_end_angle_rad)
                    theta2 = self.to_canonical_degrees(sector_start_angle_rad)
                    start_theta = theta2
                # draw gridlines
                if gridline_diameters is not None:
                    for diameter in gridline_diameters:
                        arc = mpl.patches.Arc(
                            xy=(0, 0), width=diameter, height=diameter,
                            theta1=theta1, theta2=theta2, linewidth=linewidth_grid,
                            edgecolor=color_grid, transform=ax.transData._b,
                            zorder=0, alpha=0.25,
                        )
                        ax.add_patch(arc)
                # draw gridlabels
                if (gridlabel_radii is not None) and draw_label:
                    interspace_rad = self.sector_interspace_rad[idx]
                    text_angle = sector_start_angle_rad - 0.5 * interspace_rad
                    for radius, label in zip(gridlabel_radii, gridlabels):
                        rot = self.get_rotation_upright(start_theta)
                        ax.text(text_angle, radius, label, rotation=rot, ha='center', va='center', fontsize=gridlabel_size)
                        #print(sector_start_angle_rad, radius, label, rot)

    def draw_bar(self, ax, data_df, radial_range):
        check_column_existence(data_df, ('Chromosome', 'Start', 'End', 'Color',))
        # modify data
        data_df = self.modify_data(data_df, make_radius=False)
        # remove rows where Color value is missing
        data_df = data_df.loc[np.logical_not(data_df['Color'].isnull()), :]
        # prepare drawing parameters
        height = radial_range[1] - radial_range[0]
        bottom = radial_range[0]
        # draw
        for idx, row in data_df.iterrows():
            color = row['Color']
            x = row['Start_rad']
            width = row['End_rad'] - row['Start_rad']
            ax.bar(x=x, height=height, width=width, bottom=bottom, align='edge', color=color, linewidth=0)
            
    def draw_scatter(self, ax, data_df, radial_range, ymin=None, ymax=None, size=None, alpha=None):
        check_column_existence(data_df, ('Chromosome', 'Pos0', 'Y', 'Color'))
        # modify data
        data_df['Start'] = data_df['Pos0']
        data_df['End'] = data_df['Pos0'] + 1
        data_df = self.modify_data(data_df, make_radius=True, radial_range=radial_range, ymin=ymin, ymax=ymax)
        # remove rows where Y or Color value is missing
        data_df = data_df.loc[np.logical_not(data_df['Y'].isnull()), :]
        data_df = data_df.loc[np.logical_not(data_df['Color'].isnull()), :]
        # prepare drawing parameters
        x = data_df['Start_rad']
        y = data_df['Radius']
        colors = data_df['Color']
        # draw
        ax.scatter(x=x, y=y, s=size, c=colors, alpha=alpha)

    def draw_arc(self, ax, data_df, radial_range, linewidth=None, ymin=None, ymax=None):
        check_column_existence(data_df, ('Chromosome', 'Start', 'End', 'Y', 'Color'))

        data_df = self.modify_data(data_df, make_radius=True, radial_range=radial_range, ymin=ymin, ymax=ymax)
        # remove rows where Y or Color value is missing
        data_df = data_df.loc[np.logical_not(data_df['Y'].isnull()), :]
        data_df = data_df.loc[np.logical_not(data_df['Color'].isnull()), :]

        # prepare drawing parameters
        if self.theta_direction == 1:
            theta1_list = self.to_canonical_degrees(data_df['Start_rad'])
            theta2_list = self.to_canonical_degrees(data_df['End_rad'])
        elif self.theta_direction == -1:
            theta1_list = self.to_canonical_degrees(data_df['End_rad'])
            theta2_list = self.to_canonical_degrees(data_df['Start_rad'])

        diameter_list = data_df['Radius'] * 2
        color_list = data_df['Color']

        # draw
        for theta1, theta2, diameter, color in zip(
            theta1_list, theta2_list, diameter_list, color_list
        ):
            arc = mpl.patches.Arc(
                xy=(0, 0), width=diameter, height=diameter,
                theta1=theta1, theta2=theta2, linewidth=linewidth,
                edgecolor=color, transform=ax.transData._b,
            )
            ax.add_patch(arc)

    def draw_chord(self, ax, data_df, radial_pos, linewidth=None):

        def make_patch(site1_rad, site2_rad, radial_pos, color, linewidth):
            site1_vert = (site1_rad, radial_pos)
            site2_vert = (site2_rad, radial_pos)
            control_vert = (0, 0)
            verts = [
                site1_vert,
                control_vert,
                site2_vert,
            ]
            codes = [
                mpl.path.Path.MOVETO, 
                mpl.path.Path.CURVE3, 
                mpl.path.Path.CURVE3,
            ]
            patch = mpl.patches.PathPatch(
                mpl.path.Path(verts, codes), fill=False, edgecolor=color,
                linewidth=linewidth,
            )
            return patch

        # main
        check_column_existence(data_df, ('CHROM_site1', 'POS0_site1', 'CHROM_site2', 'POS0_site2', 'Color'))
        data_df = self.modify_breakends_data(data_df)
        for idx, row in data_df.iterrows():
            patch = make_patch(row['Angle_site1_rad'], row['Angle_site2_rad'], radial_pos, row['Color'], linewidth=linewidth)
            ax.add_patch(patch)

    def arghandler_draw_gridlabel(self, arg):
        if arg is True:
            return np.repeat(True, len(self.sectorlist))
        elif arg is False:
            return np.repeat(False, len(self.sectorlist))
        else:
            if len(arg) != len(self.sectorlist):
                raise Exception(f'When "draw_gridlabel" is a sequence, its length must be equal to the number of sectors.')
            if not all(isinstance(x, int) for x in arg):
                raise Exception(f'When "draw_gridlabel" is a sequence, all of its elements must be integers.')
            result = np.repeat(False, len(self.sectorlist))
            result[list(arg)] = True
            return result

    def arghandler_sector_interspace_deg(self, arg):
        if np.isscalar(arg):
            return np.repeat(arg, len(self.sectorlist))
        else:
            if len(arg) != len(self.sectorlist):
                raise Exception(f'When "sector_interspace_deg" is a sequence, its length must be equal to the number of sectors.')
            return np.array(arg)





import os
import sys
import re
import logging
import datetime
import inspect
import itertools

import handygenome.tools as tools


LOCATION_LOGLEVELS = [logging.DEBUG]
LOCATION_EXCL_PAT = re.compile(
    '|'.join(
        (
            '^<.*>$',
            '^/tmp/ipykernel_',
            '/IPython/core/interactiveshell\.py$',
            '/IPython/core/async_helpers\.py$',
            '/ipykernel/zmqshell\.py$',
            '/ipykernel/ipkernel\.py$',
            '/ipykernel/kernelbase\.py$',
            '/asyncio/events\.py$',
            '/asyncio/base_events\.py$',
            '/tornado/platform/asyncio\.py$',
            '/ipykernel/kernelapp\.py$',
            '/traitlets/config/application\.py$',
            '/ipykernel_launcher\.py$',
        )
    )
)


def _make_logger():
    LOGGER = logging.getLogger('handygenome_logger')
    LOGGER.propagate = False
    LOGGER.setLevel(logging.DEBUG)
    
    STREAMHANDLER = logging.StreamHandler()
    LOGGER.addHandler(STREAMHANDLER)
    
    STREAMHANDLER.setLevel(logging.DEBUG)

    return LOGGER, STREAMHANDLER

LOGGER, STREAMHANDLER = _make_logger()


# timestamp

def get_datestring():
    """Returns a string like '2021-12-06 11:49:55'"""
    return str(datetime.datetime.now()).split('.')[0]


def get_timestamp():
    """Returns a string like 'KST 2021-12-06 11:51:36'"""
    dt = datetime.datetime.now().astimezone()
    return f'{str(dt.tzinfo)} {str(dt).split(".")[0]}'


def print_err(*args, stderr=True, files=None, **kwargs):
    """Args:
        stderr: (Bool) Whether to write to stderr
        files: A list of file paths to which message is written
    """
    if stderr:
        print(*args, file=sys.stderr, flush=True, **kwargs)

    if files is not None:
        for fname in files:
            with open(fname, 'a') as f:
                print(*args, file=f, flush=True, **kwargs)


def print_timestamp(*args, **kwargs):
    print_err(f'[{get_timestamp()}]', *args, **kwargs)


printlog = print_timestamp


# main logger

def log(msg, level='info', add_locstring=None, verbose_locstring=None):
    # postprocess params
    level = getattr(logging, level.upper())
    if add_locstring is None:
        add_locstring = True
    if verbose_locstring is None:
        verbose_locstring = (level == getattr(logging, 'DEBUG'))

    # main
    STREAMHANDLER.setFormatter(
        logging.Formatter(
            fmt=make_logformat(level, add_locstring=add_locstring),
            datefmt='%Z %Y-%m-%d %H:%M:%S', # KST 2022-03-23 22:12:34
        )
    )
    
    if add_locstring:
        locstring = make_locstring(verbose=verbose_locstring)
        LOGGER.log(level, msg, extra={'locstring': locstring})
    else:
        LOGGER.log(level, msg)


def log_old(msg, level='info', add_locstring=None, verbose_locstring=None):
    """Deprecated in order not to use the 'root' logger; 
    using 'root' logger makes rpy2 pacakge to emit log messages.
    """

    # postprocess params
    level = getattr(logging, level.upper())
    if add_locstring is None:
        add_locstring = True
    if verbose_locstring is None:
        verbose_locstring = (level == getattr(logging, 'DEBUG'))

    # main
    logging.basicConfig(
        format=make_logformat(level, add_locstring=add_locstring),
        datefmt='%Z %Y-%m-%d %H:%M:%S', # KST 2022-03-23 22:12:34
        level=level,
        force=True,
    )
    if add_locstring:
        locstring = make_locstring(verbose=verbose_locstring)
        logging.log(level, msg, extra={'locstring': locstring})
    else:
        logging.log(level, msg)


def make_locstring(verbose=False):
    if verbose:
        return (
            '\n'
            + '\n--> '.join(
                f'{os.path.basename(finfo.filename)}: {finfo.function} (lineno {finfo.lineno})'
                for finfo in inspect.stack()
                if LOCATION_EXCL_PAT.search(finfo.filename) is None
            )
            + '\n\n'
        )
    else:
        finfo = get_calling_frameinfo()
        return f'{os.path.basename(finfo.filename)}: {finfo.function} (lineno {finfo.lineno}) |'


def make_logformat(level, add_locstring=False):
    """Helper of 'log'"""
    if level == logging.DEBUG:
        levelcol = tools.COLORS['cyan']
    elif level == logging.INFO:
        levelcol = tools.COLORS['green']
    elif level == logging.WARNING:
        levelcol = tools.COLORS['yellow']
    elif level == logging.ERROR:
        levelcol = tools.COLORS['orange']
    elif level == logging.CRITICAL:
        levelcol = tools.COLORS['red']

    levelname = levelcol + '%(levelname)s' + tools.COLORS['end']

    if add_locstring:
        return f'[%(asctime)s.%(msecs)03d {levelname}] %(locstring)s %(message)s'
    else:
        return f'[%(asctime)s.%(msecs)03d {levelname}] %(message)s'


def get_calling_frameinfo():
    """Helper of 'log'"""
    frameinfo_groups = tuple(
        tuple(subiter) for key, subiter in
        itertools.groupby(inspect.stack(), key=(lambda x: x.filename))
    )
    assert len(frameinfo_groups) >= 2
    frameinfo = frameinfo_groups[1][0]
    return frameinfo


# line number logging

def iter_lineno_logging(line_iterator, logger, logging_lineno, msgfunc=None):
    if msgfunc is None:
        def msgfunc(NR):
            return f'Processing {NR:,}th line'

    if logging_lineno is None:
        for line in line_iterator:
            yield line
    else:
        NR = 0
        for line in line_iterator:
            NR += 1
            if NR % logging_lineno == 0:
                #logger.info(msgfunc(NR))
                log(msgfunc(NR), level='info')
            yield line


# making custom logger instance


#DEFAULT_DATEFMT = '%Z %Y-%m-%d %H:%M:%S' # KST 2022-03-23 22:12:34
#DEFAULT_LOG_FORMATTERS = {
#    'without_name': logging.Formatter(
#        fmt='[%(asctime)s %(levelname)s] %(message)s', 
#        datefmt=DEFAULT_DATEFMT,
#    ),
#    'with_name': logging.Formatter(
#        fmt='[%(asctime)s %(levelname)s] %(name)s: %(message)s', 
#        datefmt=DEFAULT_DATEFMT,
#    ),
#}
#
#def get_logger(
#    name=None, 
#    formatter=None, 
#    level='info', 
#    stderr=True, 
#    filenames=None, 
#    append=False,
#):
#    if name is None:
#        #name = str(uuid.uuid4())
#        name = __name__.split('.')[-1]
#
#    if formatter is None:
#        formatter = DEFAULT_LOG_FORMATTERS['with_name']
#
#    loglevel = getattr(logging, level.upper())
#
#    logger = logging.getLogger(name)
#    logger.setLevel(loglevel)
#    logger.propagate = False
#
#    if stderr:
#        sh = logging.StreamHandler()
#        sh.setLevel(loglevel)
#        sh.setFormatter(formatter)
#        logger.addHandler(sh)
#
#    if filenames is not None:
#        assert isinstance(filenames, (tuple, list))
#        for fname in filenames:
#            fh = logging.FileHandler(fname, mode=('a' if append else 'w'))
#            fh.setLevel(loglevel)
#            fh.setFormatter(formatter)
#            logger.addHandler(fh)
#
#    return logger
#
#
## funclogger
#
#def make_funclogger(level, name):
#    formatter = logging.Formatter(
#        #fmt=(
#        #    '[%(asctime)s.%(msecs)03d %(levelname)s] '
#        #    '%(module)s.%(funcName) line %(lineno)d: %(message)s'
#        #), 
#        fmt=(
#            '[%(asctime)s.%(msecs)s %(levelname)s] '
#            '%(module)s.%(funcName) line %(lineno)s: %(message)s'
#        ), 
#        datefmt='%Z %Y-%m-%d %H:%M:%S',
#    )
#    return get_logger(
#        level=level, 
#        name=name, 
#        formatter=formatter,
#        stderr=True, 
#        filenames=None, 
#        append=False,
#    )
#
#
#FUNCLOGGER_DEBUG = make_funclogger(level='debug', name='FUNCLOGGER_DEBUG')
#FUNCLOGGER_INFO = make_funclogger(level='info', name='FUNCLOGGER_INFO')
#
#
#def get_funclogger(verbose):
#    if verbose:
#        return FUNCLOGGER_DEBUG
#    else:
#        return FUNCLOGGER_INFO
#
#
#def printlog_funclogger(msg):
#    FUNCLOGGER_INFO.info(msg)




import os
import tempfile
import subprocess
import gzip
import collections
import re

import Bio.Align

import importlib

top_package_name = __name__.split(".")[0]
common = importlib.import_module(".".join([top_package_name, "common"]))


#ALIGNER_FILLED = Bio.Align.PairwiseAligner(
#    mode='global',
#    match_score=2,
#    mismatch_score=-3,
#    query_internal_open_gap_score=-7,
#    query_internal_extend_gap_score=-2,
#    target_internal_open_gap_score=-7,
#    target_internal_extend_gap_score=-2,
#    query_left_open_gap_score=-7,
#    query_left_extend_gap_score=-2,
#)
#
#ALIGNER_UNFILLED = Bio.Align.PairwiseAligner(
#    mode='global',
#    match_score=2,
#    mismatch_score=-3,
#    query_internal_open_gap_score=-7,
#    query_internal_extend_gap_score=-2,
#    target_internal_open_gap_score=-7,
#    target_internal_extend_gap_score=-2,
#)


# parameters are derived from Biopython documentation
ALIGNER_BLASTN = Bio.Align.PairwiseAligner(
    match_score=2,
    mismatch_score=-3,
    query_internal_open_gap_score=-7,
    query_internal_extend_gap_score=-2,
    target_internal_open_gap_score=-7,
    target_internal_extend_gap_score=-2,
)


def get_aligner_1():
    aligner = Bio.Align.PairwiseAligner(
        match_score=2,
        mismatch_score=-3,
        query_internal_open_gap_score=-7,
        query_internal_extend_gap_score=-2,
        target_internal_open_gap_score=-7,
        target_internal_extend_gap_score=-2,
    )

    return aligner


def get_aligner_2():
    aligner = Bio.Align.PairwiseAligner(
        match_score=2,
        mismatch_score=-3,
        query_internal_open_gap_score=-7,
        query_internal_extend_gap_score=-1,
        target_internal_open_gap_score=-7,
        target_internal_extend_gap_score=-1,
    )

    return aligner


def do_align(query, target, aligner):
    """Arguments:
    query, target: str object
    """
    return aligner.align(target, query)


def get_idx_tuples(alignment):
    idx_tuples = list()
    for idx in range(1, len(alignment.path)):
        idx_tuples.append( tuple(zip(alignment.path[idx-1], alignment.path[idx])) )

    return idx_tuples


def get_cigartuples(idx_tuples):
    cigartuples_list = list()

    for idx, tup_pair in enumerate(idx_tuples):
        target_len = tup_pair[0][1] - tup_pair[0][0]
        query_len = tup_pair[1][1] - tup_pair[1][0]

        if target_len == 0 and query_len != 0: # insertion
            cigartuples_list.append((1, query_len))
        elif target_len != 0 and query_len == 0: # deletion
            if idx != 0 and idx != len(idx_tuples)-1:
                cigartuples_list.append((2, target_len))
        else: # match
            cigartuples_list.append((0, target_len))

    return tuple(cigartuples_list)



import pysam

import handygenome.refgenome.refgenome as refgenome
from handygenome.variant.vcfspec import Vcfspec
import handygenome.variant.vcfspec as libvcfspec
import handygenome.vcfeditor.initvcf as initvcf
import handygenome.vcfeditor.misc as vcfmisc
import handygenome.vcfeditor.indexing as indexing
from handygenome.annotation.callerinfo import CallerInfo
from handygenome.variant.vcfspec import VcfspecComponents


PASS_SET = {'PASS'}


def sanity_check_pre_split(vr):
    # sanity check 1
    if len(vr.alts) != 1:
        raise Exception(f'Comma-separated ALT detected:\n{vr}')
    if '*' in vr.ref or '*' in vr.alts[0]:
        raise Exception(f'"*" included in REF or ALT:\n{vr}')


def sanity_check_post_split(vr, REFsp, ALTsp):
    if len(REFsp) > 1 and len(ALTsp) > 1:
        raise Exception(f'Both REF and ALT include "/":\n{vr}')

    if len(REFsp) > 2 or len(ALTsp) > 2:
        raise Exception(f'REF or ALT have more than two "/"-separated fields:\n{vr}')

    if len(ALTsp) > 1:
        # 1    937795  .   C   CATTT/ATTTT
        if len(vr.ref) > 1:
            raise Exception(f'ALT is multi-field but REF is not a single-character:\n{vr}')

        if all(len(x) == 1 for x in ALTsp):  # snps
            pass
        else:  # indels
            if not (
                ALTsp[0][0] == vr.ref and
                len(ALTsp[0]) > 1
            ):
                raise Exception(f'Unexpected multi-ALT pattern:\n{vr}')
                #raise Exception(f'ALT is multi-field but the first character of ALT is not the same as REF:\n{vr}')

    if len(REFsp) > 1:
        # 1    1029791 .   CT/+T   C
        if len(vr.alts[0]) > 1:
            raise Exception(f'REF is multi-field but ALT is not a single-character:\n{vr}')
        if not (
            REFsp[0][0] == vr.alts[0] and
            len(REFsp[0]) > 1
        ):
            raise Exception(f'Unexpected multi-REF pattern:\n{vr}')
        #if REFsp[0][0] != vr.alts[0]:
        #    raise Exception(f'REF is multi-field but the first character of REF is not the same as ALT:\n{vr}')


def handle_multi_ALT_indel(vr, fasta, REFsp, ALTsp):
    # Now REF is composed of a single letter
    # First element of ALTsp starts with a letter the same as REF. Strip it.
        # examples:
        # 1    937795  .   C   CATTT/ATTTT
        # 1    51864   .   C   CA/-A
    assert len(ALTsp[0]) > 1
    ALTsp[0] = ALTsp[0][1:]
    merge_components = list()

    for ALT_unit in ALTsp:
        if ALT_unit.startswith('-'):
            # It is postulated to be a deletion
            ALT_unit = ALT_unit[1:]  # strip leading "-"
            pos0 = vr.pos - 1
            fetched_seq_nextto_ref = fasta.fetch(vr.contig, pos0 + 1, pos0 + 1 + len(ALT_unit))
            if fetched_seq_nextto_ref != ALT_unit:
                raise Exception(
                    f'ALT string after "-" character is different from fetched reference sequence:\n'
                    f'Fetched reference sequence: {fetched_seq_nextto_ref} (chrom={vr.contig}, start0={pos0 + 1}, end0={pos0 + 1 + len(ALT_unit)})\n'
                    f'vr:\n{vr}'
                )
            ref = vr.ref + fetched_seq_nextto_ref
            alt = vr.ref
        else:
            # Postulated to be an insertion
            ref = vr.ref
            alt = vr.ref + ALT_unit

        merge_components.append(Vcfspec(chrom=vr.contig, pos=vr.pos, ref=ref, alts=(alt,), fasta=fasta))

    return libvcfspec.merge(*merge_components)


def handle_multi_REF(vr, fasta, REFsp, ALTsp):
    # Now REF is composed of a single letter
    # First element of REFsp starts with a letter the same as ALT. Strip it.
        # examples:
        # 1    1092725 .   CCCATCCCCGCCATCCCCG/+CCCG   C
        # 1    1029791 .   CT/+T   C
        # 1    1204224 .   ATT/T   A
    assert len(REFsp[0]) > 1
    REFsp[0] = REFsp[0][1:]
    merge_components = list()

    for REF_unit in REFsp:
        if REF_unit.startswith('+'):
            # It is postulated to be an insertion
            REF_unit = REF_unit[1:]
            ref = vr.alts[0]
            alt = vr.alts[0] + REF_unit
        else:
            # It is postulated to be a deletion
            ref = vr.alts[0] + REF_unit
            alt = vr.alts[0]

        merge_components.append(Vcfspec(chrom=vr.contig, pos=vr.pos, ref=ref, alts=(alt,), fasta=fasta))

    return libvcfspec.merge(*merge_components)


def extract_vcfspec(vr, fasta):
    sanity_check_pre_split(vr)

    REFsp = vr.ref.split('/')
    ALTsp = vr.alts[0].split('/')

    sanity_check_post_split(vr, REFsp, ALTsp)

    # main
    if len(REFsp) == 1 and len(ALTsp) == 1:
        return Vcfspec(chrom=vr.contig, pos=vr.pos, ref=vr.ref, alts=vr.alts, fasta=fasta)
    elif len(REFsp) == 1 and len(ALTsp) > 1:
        if all(len(x) == 1 for x in ALTsp):  # snps
            return Vcfspec(chrom=vr.contig, pos=vr.pos, ref=vr.ref, alts=ALTsp, fasta=fasta)
        else:
            return handle_multi_ALT_indel(vr, fasta, REFsp, ALTsp)
    elif len(REFsp) > 1 and len(ALTsp) == 1:
        return handle_multi_REF(vr, fasta, REFsp, ALTsp)


def write_outfile(vcfspec_list, outfile_path, output_header, mode, caller_info, index):
    with pysam.VariantFile(outfile_path, header=output_header, mode=mode) as out_vcf:
        for vcfspec in vcfspec_list:
            new_vr = output_header.new_record()
            vcfspec.apply_to_vr(new_vr)
            caller_info.write(new_vr)
            out_vcf.write(new_vr)

    if index:
        indexing.index_vcf(outfile_path)


def modify_varscan_vcf(
    infile_path_list, 
    somatic_outfile_path, 
    germline_outfile_path=None,
    refver=None, 
    fasta=None,
    mode_bcftools='z',
    mode_pysam=None, 
    pass_only=True,
    index=True,
):
    """Merge all files in "infile_path_list" to make a single sorted output.
    INFO/FORMAT contents are all removed.
    Args:
        germline_outfile_path: If not set, germline variant records are not written.
    """
    # sanity check
    if (
        (refver is None) and
        (fasta is None)
    ):
        raise Exception(f'At least one of "refver" or "fasta" must be set.')

    # set params
    if fasta is None:
        fasta = refgenome.get_fasta(refver)
    chromdict = refgenome.ChromDict.from_fasta(fasta)

    output_header = initvcf.create_header(chromdict=chromdict)
    CallerInfo.add_meta(output_header)
    VcfspecComponents.add_meta(output_header)

    caller_info = CallerInfo.from_caller_names(['VarScan2'])

    mode = vcfmisc.write_mode_arghandler(mode_bcftools, mode_pysam)
    somatic_vcfspec_list = list()
    germline_vcfspec_list = list()

    # collect vcfspecs
    for infile_path in infile_path_list:
        with pysam.VariantFile(infile_path) as in_vcf:
            for vr in in_vcf.fetch():
                if pass_only:
                    if set(vr.filter) != PASS_SET:
                        continue

                vcfspec = extract_vcfspec(vr, fasta)
                if vcfspec.chrom not in chromdict.contigs:
                    continue

                if vr.info['SOMATIC']:
                    somatic_vcfspec_list.append(vcfspec)
                else:
                    if germline_outfile_path is not None:
                        germline_vcfspec_list.append(vcfspec)

    # sort vcfspecs
    #sortkey = common.get_vcfspec_sortkey(chromdict)
    #somatic_vcfspec_list.sort(key=sortkey)
    somatic_vcfspec_list.sort(key=(lambda x: x.get_sortkey(chromdict)))
    #germline_vcfspec_list.sort(key=sortkey)
    germline_vcfspec_list.sort(key=(lambda x: x.get_sortkey(chromdict)))

    # write output
    write_outfile(somatic_vcfspec_list, somatic_outfile_path, output_header, mode, caller_info, index)
    if germline_outfile_path is not None:
        write_outfile(germline_vcfspec_list, germline_outfile_path, output_header, mode, caller_info, index)





import inspect
import os
import collections
import functools

import numpy as np
import pandas as pd

import handygenome.tools as tools
import handygenome.deco as deco
import handygenome.refgenome.refgenome as refgenome
from handygenome.refgenome.refgenome import NoXYError
import handygenome.refgenome.refverfile as refverfile
import handygenome.genomedf.genomedf as libgenomedf
from handygenome.genomedf.genomedf import GenomeDataFrame as GDF


###############
# terminology #
###############

'''
CNg: total copy number in germline
CNt: total copy number of the target (tumor) sample
Bg: B allele copy number in germline
Bt: B allele copy number of the target (tumor) sample
'''


#############
# constants #
#############

DEFAULT_CNG_COLNAME = 'default_CNg'
DEFAULT_BG_COLNAME = 'default_Bg'
MALE_HAPLOID_CHROMS = ('X', 'Y', 'chrX', 'chrY')
DEFAULT_CNT_WEIGHT = 5


###############
# CCF classes #
###############

class CCFInfo(
    collections.namedtuple('CCFInfo', ('CNm', 'ccf', 'mutated_allele'))
):
    pass


class CPPair(
    collections.namedtuple('CPPair', ('cellularity', 'ploidy'))
):
    pass


##############
# exceptions #
##############

class DefaultCNgUnavailableError(Exception):
    pass


##############
# decorators #
##############

def deco_cellularity_sanitycheck(func):
    sig = inspect.signature(func)
    if 'cellularity' not in sig.parameters.keys():
        raise Exception(f'"cellularity" is not included in the paramter list.')
    
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        ba = sig.bind(*args, **kwargs)
        ba.apply_defaults()
        if ba.arguments['cellularity'] == 0:
            raise Exception(f'"cellularity" must not be 0.')

        return func(*ba.args, **ba.kwargs)

    return wrapper


################
# miscellanous #
################

def check_haploid(is_female, chrom):
    return (
        (not is_female) 
        and refgenome.check_xy_chrom(chrom)
    )


############################
# germline sample solution #
############################

def get_default_CNg_Bg_diploid(refver, is_female):
    """Assumptions
        1) X, Y is sex chromosome
        2) Male has one copy of each sex chromosomes
        3) pseudoautosomal region is available
    """
    chromdict = refgenome.get_chromdict(refver)
    autosomal_chroms = [
        x for x in chromdict.contigs 
        if (
            refgenome.check_assembled_chrom(x)
            and (not refgenome.check_xy_chrom(x))
        )
    ]  # 1, 2, ...,
    try:
        X_chrom, Y_chrom = chromdict.XY_names
    except NoXYError:
        raise DefaultCNgUnavailableError(f'Default germline CN information is not available for this reference version.')

    par_gdf = refverfile.get_par_gdf(refver)

    # autosomal
    all_chrom_gdf = GDF.all_regions(refver, assembled_only=True)
    autosomal_gdf = all_chrom_gdf.subset_chroms(autosomal_chroms)
    autosomal_gdf[DEFAULT_CNG_COLNAME] = float(2)
    autosomal_gdf[DEFAULT_BG_COLNAME] = float(1)

    # sex
    if is_female:
        sex_gdf = all_chrom_gdf.subset_chroms([X_chrom, Y_chrom])
        is_X = (sex_gdf.chroms == X_chrom)
        sex_gdf.loc[is_X, DEFAULT_CNG_COLNAME] = float(2)
        sex_gdf.loc[is_X, DEFAULT_BG_COLNAME] = float(1)
        sex_gdf.loc[~is_X, DEFAULT_CNG_COLNAME] = float(0)
        sex_gdf.loc[~is_X, DEFAULT_BG_COLNAME] = np.nan
    else:
        all_sex_gdf = all_chrom_gdf.subset_chroms([X_chrom, Y_chrom])
        nonpar_sex_gdf = all_sex_gdf.subtract(par_gdf)
        nonpar_sex_gdf[DEFAULT_CNG_COLNAME] = float(1)
        nonpar_sex_gdf[DEFAULT_BG_COLNAME] = np.nan

        par_sex_gdf = par_gdf.copy()
        is_X = (par_sex_gdf['Chromosome'] == X_chrom)
        par_sex_gdf.loc[is_X, DEFAULT_CNG_COLNAME] = float(2)
        par_sex_gdf.loc[is_X, DEFAULT_BG_COLNAME] = float(1)
        par_sex_gdf.loc[~is_X, DEFAULT_CNG_COLNAME] = float(0)
        par_sex_gdf.loc[~is_X, DEFAULT_BG_COLNAME] = np.nan

        sex_gdf = GDF.concat([nonpar_sex_gdf, par_sex_gdf])

    # result
    result = GDF.concat([autosomal_gdf, sex_gdf])
    result.sort()
    return result


def clonal_solution_from_germline(cnv_seg_gdf, is_female, ploidy):
    # add default CNg and Bg
    cnv_seg_gdf_copy = cnv_seg_gdf.copy()

    # find clonal CNt
    onecopy_depth = cnv_seg_gdf_copy.find_onecopy_depth(is_female, ploidy)
    K = get_K_from_onecopy_depth(onecopy_depth)
    clonal_CNt = find_clonal_CNt(
        corrected_depth=cnv_seg_gdf_copy.norm_depth_mean, 
        cellularity=1, 
        K=K,
        CNg=None, 
    )

    # find clonal B
    B_isnan_selector = np.isnan(cnv_seg_gdf_copy.get_default_Bg(is_female))
    B_notnan_selector = ~B_isnan_selector

    clonal_Bt = np.empty(cnv_seg_gdf_copy.nrow, dtype=float)
    clonal_Bt[B_isnan_selector] = np.nan

    relevant_bafs = cnv_seg_gdf_copy.get_corrected_baf()[B_notnan_selector]
    assert not np.isnan(relevant_bafs).any()
    clonal_Bt[B_notnan_selector] = find_clonal_Bt(
        baf=relevant_bafs, 
        CNt=clonal_CNt[B_notnan_selector], 
        cellularity=1, 
        CNg=None, 
        Bg=None,
    )

    return clonal_CNt, clonal_Bt


def solution_from_targetsample():
    average_CNt = depth_seg_gdf.norm_depth / onecopy_depth
    clonal_CNt = np.rint(average_CNt)

    # 4. determine regions where clonal solutions are unsatisfactory
    depth_diffs = np.abs((onecopy_depth * clonal_CNt) - depth_seg_gdf.norm_depth)
    diff_thresholds = (onecopy_depth * threshold_factor) * average_CNt
        # Where "average_CNt" is 2, "diff_thresholds" value becomes:
        # 2 * onecopy_depth * threshold_factor
    unfit_selector = depth_diffs > diff_thresholds
    unfit_selector_idxs = np.nonzero(unfit_selector)[0]

    # 5. find subclonal solutions
    unfit_average_CNt = average_CNt[unfit_selector_idxs]
    unfit_clonal_CNt = depth_seg_gdf['default_CNg'][unfit_selector_idxs]

    unfit_clonal_CNt, unfit_subclonal_CNt = get_CN_candidates_type2(unfit_average_CNt, clonal_CNt=unfit_clonal_CNt)

    return depth_seg_gdf


@deco_cellularity_sanitycheck
def find_clonal_CNt(corrected_depth, cellularity, K, CNg=None):
    """Assume 'corrected_depth' is corrected so that position-dependent bias has been removed.
    It can be formulated like:
        'corrected_depth' = K * (CNg * (1 - cellularity) + CNt * cellularity)

    derivation:
        corrected_depth = K * (CNg * (1 - cellularity) + CNt * cellularity)
        corrected_depth / K = CNg * (1 - cellularity) + CNt * cellularity
        (corrected_depth / K) - (CNg * (1 - cellularity)) = CNt * cellularity
        CNt = ( (corrected_depth / K) - (CNg * (1 - cellularity)) ) / cellularity
    """
    if (cellularity != 1) and (CNg is None):
        raise Exception(f'When cellularity is not 1, "CNg" must be given.')

    if cellularity == 1:
        CNt = (corrected_depth / K)
    else:
        CNt = (
            (corrected_depth / K) 
            - (CNg * (1 - cellularity))
        ) / cellularity

    CNt = np.rint(CNt)
    return CNt


@deco_cellularity_sanitycheck
def find_clonal_Bt(baf, CNt, cellularity, CNg=None, Bg=None):
    """
    derivation:
        baf = (
            (Bt * cellularity + Bg * (1 - cellularity))
            / (CNt * cellularity + CNg * (1 - cellularity))
        )
        Bt * cellularity + Bg * (1 - cellularity) = (
            baf * (CNt * cellularity + CNg * (1 - cellularity))
        )
        Bt * cellularity = (
            (baf * (CNt * cellularity + CNg * (1 - cellularity)))
            - (Bg * (1 - cellularity))
        )
        Bt = (
            (baf * (CNt * cellularity + CNg * (1 - cellularity)))
            - (Bg * (1 - cellularity))
        ) / cellularity
        Bt = (
            (baf * (CNt + CNg * (1 / cellularity - 1)))
            - (Bg * (1 / cellularity - 1))
        )
    """

    if (
        (cellularity != 1) 
        and (
            (CNg is None)
            or (Bg is None)
        )
    ):
        raise Exception(f'When cellularity is not 1, "CNg" and "Bg" must be given.')

    c = (1 / cellularity) - 1
    if c == 0:
        Bt = baf * CNt 
    else:
        Bt = (baf * (CNt + CNg * c)) - (Bg * c)

    Bt = np.rint(Bt)
    Bt = np.clip(Bt, 0, np.floor(CNt / 2))
    return Bt


def get_K_from_onecopy_depth(onecopy_depth):
    """Only make sense with germline wgs sample

    derivation:
        corrected_depths = K * (CNg * (1 - cellularity) + CNt * cellularity)
        corrected_depths = K * (CNt)  (since cellularity == 1)
        K = corrected_depths / CNt
        K = onecopy_depth / 1
    """
    return onecopy_depth


#####################################
# helpers - CN candidate generators #
#####################################

@deco.get_deco_atleast1d(['average_CNt'])
def get_CN_candidates_type1(average_CNt, *, min_candidate_num=5, candidate_num_fraction=0.5):
    """lower and upper candidates are given
    """
    assert average_CNt.ndim == 1

    num_candidates, max_dimsize, CN_shifts = get_CN_candidate_shifts(
        average_CNt, min_candidate_num, candidate_num_fraction
    )
    upperCN_borders, lowerCN_borders = tools.get_nearest_integer_bounds(average_CNt, elevate=True)

    upper_CNt = upperCN_borders[:, np.newaxis] + CN_shifts
    lower_CNt = lowerCN_borders[:, np.newaxis] - CN_shifts
    lower_CNt[lower_CNt < 0] = np.nan

    return lower_CNt, upper_CNt


@deco.get_deco_squeeze_atleast1d(['average_CNt'])
def get_CN_candidates_type2(average_CNt, *, clonal_CNt=None, min_candidate_num=5, candidate_num_fraction=0.5):
    """Candidates are labeled as clonal or subclonal. Clonal values can be optionally given.
    """
    assert average_CNt.ndim == 1

    num_candidates, max_dimsize, CN_shifts = get_CN_candidate_shifts(
        average_CNt, min_candidate_num, candidate_num_fraction
    )

    if clonal_CNt is None:
        clonal_CNt = np.rint(average_CNt)
    else:
        assert average_CNt.shape == clonal_CNt.shape

    gt_clonal_selector = (average_CNt > clonal_CNt)

    gt_clonal_subCNt = clonal_CNt[gt_clonal_selector][:, np.newaxis] + CN_shifts[gt_clonal_selector]
    lt_clonal_subCNt = clonal_CNt[~gt_clonal_selector][:, np.newaxis] - CN_shifts[~gt_clonal_selector]
    lt_clonal_subCNt[lt_clonal_subCNt < 0] = np.nan

    subclonal_CNt = np.empty((len(average_CNt), max_dimsize), dtype=float)
    subclonal_CNt[gt_clonal_selector, :] = gt_clonal_subCNt
    subclonal_CNt[~gt_clonal_selector, :] = lt_clonal_subCNt

    return clonal_CNt, subclonal_CNt


def get_CN_candidate_shifts(average_CNt, min_candidate_num, candidate_num_fraction):
    num_candidates = np.maximum(
        min_candidate_num, 
        np.rint(average_CNt * candidate_num_fraction),
    )
    max_dimsize = np.nanmax(num_candidates, axis=None).astype(int)

    CN_shifts = np.tile(
        np.arange(max_dimsize)[np.newaxis, :], 
        (len(average_CNt), 1),
    )
    CN_shifts = CN_shifts.astype(float)
    nan_selector = np.stack(
        [
            (
                np.repeat(True, max_dimsize)
                if np.isnan(x) else
                np.repeat([False, True], (x, max_dimsize - x)) 
            )
            for x in num_candidates
        ], 
        axis=0,
    )
    CN_shifts[nan_selector] = np.nan

    return num_candidates, max_dimsize, CN_shifts


#############################
# helpers - B determination #
#############################

def get_B_from_clonal_CN(clonal_CNt, bafs):
    pass
    






import os
import re
import argparse
import gzip

import pysam

import handygenome.network as network
import handygenome.tools as tools
import handygenome.workflow as workflow
import handygenome.workflow.toolsetup as toolsetup
import handygenome.refgenome.refgenome as refgenome


DEFAULT_OUTFILE_BASENAME = 'ensembl_geneset_sorted.gff3.gz'
URL_GRCH37 = 'http://ftp.ensembl.org/pub/grch37/release-105/gff3/homo_sapiens/Homo_sapiens.GRCh37.87.gff3.gz'
URL_GRCH38 = 'http://ftp.ensembl.org/pub/current_gff3/homo_sapiens/Homo_sapiens.GRCh38.105.gff3.gz'
DEFAULT_OUTFILE_PATH_GRCH37 = os.path.join(os.getcwd(), re.sub('gff3.gz$', 'grch37.gff3.gz', DEFAULT_OUTFILE_BASENAME))
DEFAULT_OUTFILE_PATH_GRCH38 = os.path.join(os.getcwd(), re.sub('gff3.gz$', 'grch38.gff3.gz', DEFAULT_OUTFILE_BASENAME))


def argument_parser(cmdargs):
    parser_dict = workflow.init_parser(
        description=(
            f'The original gff3 gene set files are downloaded, filtered, '
            f'sorted, bgzipped, and tabix-indexed. These lines are filtered '
            f'out: containing "###" or "feature" value is "chromosome" '
            f'or "supercontig".'))

    parser_dict['optional'].add_argument(
        '--outfile-grch37', dest='outfile_path_grch37', 
        default=DEFAULT_OUTFILE_PATH_GRCH37,
        required=False, help=f'Output gff3 file path for grch37')

    parser_dict['optional'].add_argument(
        '--outfile-grch38', dest='outfile_path_grch38', 
        default=DEFAULT_OUTFILE_PATH_GRCH38,
        required=False, help=f'Output gff3 file path for grch38')

    parser_dict['optional'].add_argument(
        '--downloaded-file-grch37', dest='download_path_grch37', 
        required=False, 
        help=(f'(Optional) previously downloaded original gene set gff3 '
              f'file path for grch37'))

    parser_dict['optional'].add_argument(
        '--downloaded-file-grch38', dest='download_path_grch38', 
        required=False, 
        help=(f'(Optional) previously downloaded original gene set gff3 '
              f'file path for grch38'))

    workflow.add_logging_args(parser_dict)

    args = parser_dict['main'].parse_args(cmdargs)

    return args


def download(logger, url, download_path):
    logger.info(f'Beginning download of url "{url}" to "{download_path}"')
    network.download(url, download_path)
    logger.info(f'Download finished')


def get_line_lists(download_path, chromdict, asmblspec, output_chrname_version):
    line_list_comment = list()
    linesp_list = list()
    with gzip.open(download_path, 'rt') as infile:
        for line in infile:
            linestrip = tools.rm_newline(line)
            if linestrip == '###':
                continue
            else:
                if linestrip.startswith('#'):
                    line_list_comment.append(linestrip)
                else:
                    linesp = linestrip.split('\t')
                    if linesp[2] in ('chromosome', 'supercontig'):
                        continue
                    if linesp[0] not in chromdict:
                        linesp[0] = asmblspec.convert(
                            linesp[0], output_chrname_version)

                    linesp_list.append(linesp)

    return line_list_comment, linesp_list


def sort_linesp_list(chromdict, linesp_list):
    def sortkey(linesp):
        return tools.coord_sortkey(linesp[0], int(linesp[3]), chromdict)
    linesp_list_sorted = sorted(linesp_list, key = sortkey)

    return linesp_list_sorted


def write_uncompressed_file(outfile_path_uncomp, line_list_comment, 
                            linesp_list_sorted): 
    with open(outfile_path_uncomp, 'wt') as outfile:
        for linestrip in line_list_comment:
            outfile.write(linestrip + '\n')
        for linesp in linesp_list_sorted:
            outfile.write('\t'.join(linesp) + '\n')


def convert_common(original_url, output_chrname_version, refver, logger, 
                   outfile_path, download_path=None):
    # download
    if download_path is None:
        download_path = workflow.get_tmpfile_path()
        download(logger, original_url, download_path)
        rm_downloaded = True
    else:
        rm_downloaded = False

    # set refgenome dbs
    asmblspec = refgenome.SPECS[refver]
    chromdict = asmblspec.chromdicts[output_chrname_version]

    # load
    logger.info(f'Loading the original file')
    line_list_comment, linesp_list = get_line_lists(download_path, chromdict, 
                                                    asmblspec,
                                                    output_chrname_version)

    # sort
    logger.info(f'Sorting the original file lines')
    linesp_list_sorted = sort_linesp_list(chromdict, linesp_list)

    # write uncompressed outfile
    logger.info(f'Writing the sorted lines')
    outfile_path_uncomp = outfile_path + '.uncompressed'
    write_uncompressed_file(outfile_path_uncomp, line_list_comment, linesp_list_sorted)

    # compress and index
    logger.info(f'Compressing with bgzip and indexing with tabix')
    pysam.tabix_compress(outfile_path_uncomp, outfile_path)
    pysam.tabix_index(outfile_path, preset = 'gff')

    # remove temp files
    os.remove(outfile_path_uncomp)
    if rm_downloaded:
        os.remove(download_path)


def convert_grch37(logger, outfile_path, download_path=None):
    convert_common(URL_GRCH37, 'nochr_plus_genbank', 'hg19', logger, 
                   outfile_path, download_path)


def convert_grch38(logger, outfile_path, download_path=None):
    convert_common(URL_GRCH38, 'ucsc', 'hg38', logger, outfile_path, 
                   download_path)


def main(cmdargs):
    args = argument_parser(cmdargs)
    logger = toolsetup.setup_logger(args)

    convert_grch37(logger, args.outfile_path_grch37, args.download_path_grch37)
    convert_grch38(logger, args.outfile_path_grch38, args.download_path_grch38)
    
    logger.info('ALL SUCCESSFULLY FINISHED')



import os
import itertools
import functools

import numpy as np
import pandas as pd

import handygenome
import handygenome.network as network
import handygenome.deco as deco


COLORS_SBS6 = {
    'C>A': np.array([3, 189, 239]) / 256,
    'C>G': np.array([1, 1, 1]) / 256,
    'C>T': np.array([228, 41, 38]) / 256,
    'T>A': np.array([203, 202, 202]) / 256,
    'T>C': np.array([162, 207, 99]) / 256, 
    'T>G': np.array([236, 199, 197]) / 256,
    'other': 'y',
}

AVAILABLE_REFVERS = ('GRCh38', 'GRCh37', 'mm10')
AVAILABLE_CAT_TYPES = ('sbs96', 'id83', 'dbs78', 'cn48')

DATA_DIR = os.path.join(handygenome.DATADIR, 'signature_data')
if not os.path.exists(DATA_DIR):
    os.mkdir(DATA_DIR)

# SIGNATURE_DATA_URLS
SIGNATURE_DATA_URLS = dict()
SIGNATURE_DATA_URLS['id83'] = 'https://cancer.sanger.ac.uk/signatures/documents/1907/COSMIC_v3.3_ID_GRCh37.txt'
SIGNATURE_DATA_URLS['cn48'] = 'https://cancer.sanger.ac.uk/signatures/documents/2044/COSMIC_v3.3_CN_GRCh37.txt'
SIGNATURE_DATA_URLS['GRCh37'] = {
    'sbs96': 'https://cancer.sanger.ac.uk/signatures/documents/1908/COSMIC_v3.3_SBS_GRCh37.txt',
    'id83': SIGNATURE_DATA_URLS['id83'],
    'dbs78': 'https://cancer.sanger.ac.uk/signatures/documents/1902/COSMIC_v3.3_DBS_GRCh37.txt',
    'cn48': SIGNATURE_DATA_URLS['cn48'],
    }
SIGNATURE_DATA_URLS['GRCh38'] = {
    'sbs96': 'https://cancer.sanger.ac.uk/signatures/documents/1909/COSMIC_v3.3_SBS_GRCh38.txt',
    'id83': SIGNATURE_DATA_URLS['id83'],
    'dbs78': 'https://cancer.sanger.ac.uk/signatures/documents/1903/COSMIC_v3.3_DBS_GRCh38.txt',
    'cn48': SIGNATURE_DATA_URLS['cn48'],
    }
SIGNATURE_DATA_URLS['mm10'] = {
    'sbs96': 'https://cancer.sanger.ac.uk/signatures/documents/1911/COSMIC_v3.3_SBS_mm10.txt',
    'id83': SIGNATURE_DATA_URLS['id83'],
    'dbs78': 'https://cancer.sanger.ac.uk/signatures/documents/1905/COSMIC_v3.3_DBS_mm10.txt',
    'cn48': SIGNATURE_DATA_URLS['cn48'],
    }
###

# SIGNATURE_DATA_FILE_PATHS
SIGNATURE_DATA_FILE_PATHS = dict()
for refver in AVAILABLE_REFVERS:
    subdict = dict()
    for catalogue_type in AVAILABLE_CAT_TYPES:
        url = SIGNATURE_DATA_URLS[refver][catalogue_type]
        path = os.path.join(DATA_DIR, os.path.basename(url))
        subdict[catalogue_type] = path
    SIGNATURE_DATA_FILE_PATHS[refver] = subdict
###


def get_cossim(v1, v2):
    if not (np.ndim(v1) == np.ndim(v2) == 1 and
            np.shape(v1) == np.shape(v2)):
        raise Exception(f'Each input argument must be equivalent to '
                        f'a one-dimensional array with the same length.')

    v1_2norm = np.linalg.norm(v1, ord=2, axis=None)
    v2_2norm = np.linalg.norm(v2, ord=2, axis=None)
    if v1_2norm == 0 or v2_2norm == 0:
        return np.nan
    else:
        return min(1.0, np.dot(v1, v2) / (v1_2norm * v2_2norm))


def create_catalogue_keys_sbs6(as_tuple=False):
    result = list()
    for ref in 'CT':
        alts = sorted(set('ACGT') - set(ref))
        for alt in alts:
            if as_tuple:
                item = (ref, alt)
            else:
                item = f'{ref}>{alt}'
            result.append(item)

    return tuple(result)


def create_catalogue_keys_sbs96(as_tuple=False):
    result = list()
    for ref in 'CT':
        alts = sorted(set('ACGT') - set(ref))
        for alt in alts:
            for pre, post in itertools.product('ACGT', repeat=2):
                if as_tuple:
                    item = (pre, ref, alt, post)
                else:
                    item = f'{pre}[{ref}>{alt}]{post}'
                result.append(item)

    return tuple(result)


@deco.get_deco_arg_choices({'refver': AVAILABLE_REFVERS})
@deco.get_deco_arg_choices({'catalogue_type': AVAILABLE_CAT_TYPES})
def download_signature_data(refver, catalogue_type):
    """Args:
        catalogue_type: Mutation type (e.g. sbs, id, dbs, cn)
    """
    data_url = SIGNATURE_DATA_URLS[refver][catalogue_type]
    data_file_path = SIGNATURE_DATA_FILE_PATHS[refver][catalogue_type]
    network.download(data_url, data_file_path)
    

@deco.get_deco_arg_choices({'refver': AVAILABLE_REFVERS})
@deco.get_deco_arg_choices({'catalogue_type': AVAILABLE_CAT_TYPES})
@functools.cache
def load_signature_data(refver, catalogue_type):
    """num_rows:
        sbs: 96
        id: 83
        dbs: 78
        cn: 48
    """
    data_file_path = SIGNATURE_DATA_FILE_PATHS[refver][catalogue_type]
    if not os.path.exists(data_file_path):
        download_signature_data(refver, catalogue_type)
    sigdata = pd.read_csv(data_file_path, sep='\t', index_col=0)

    return sigdata


@functools.cache
def get_catalogue_keys(catalogue_type):
    if catalogue_type in ('sbs96', 'id83', 'dbs78', 'cn48'):
        sigdata = load_signature_data('GRCh37', catalogue_type)
        return tuple(sigdata.index)
    elif catalogue_type == 'sbs6':
        return create_catalogue_keys_sbs6(as_tuple=False)
    else:
        raise Exception(f'Unavailable catalogue type')


def get_catalogue_type(pd_index):
    pd_index_set = set(pd_index)
    okay_cattypes = list()
    for catalogue_type in ('sbs96', 'id83', 'dbs78', 'cn48', 'sbs6'):
        if pd_index_set == set(get_catalogue_keys(catalogue_type)):
            okay_cattypes.append(catalogue_type)

    if len(okay_cattypes) == 1:
        return okay_cattypes[0]
    else:
        raise Exception(f'Unknown catalogue type')
        



import os
import sys
import re
import subprocess
import multiprocessing
import tempfile

import sequenza.commands

import handygenome.common as common
import handygenome.assemblyspec as assemblyspec


DEFAULT_GCWIG_PATHS = assemblyspec.RefverDict(
    {
        'GRCh37': '/home/users/data/01_reference/human_g1k_v37/human_g1k_v37.gc50Base.txt.gz',
        'GRCh38': '/home/users/pjh/References/reference_genome/GRCh38/GCA_for_alignment_pipelines/no_alt_plus_hs38d1/GCA_000001405.15_GRCh38_no_alt_plus_hs38d1_analysis_set.gc50Base.txt.gz',
    }
)
DEFAULT_HOM = 0.9
DEFAULT_HET = 0.25


def run_sequenza_utils(args):
    sys.argv = ['sequenza_utils'] + args
    try:
        result = sequenza.commands.main()
    except SystemExit:
        pass
    #sys.exit(result)


def run_sequenza_utils_subprocess(args):
    pass
    #p = subprocess.run


def fasta_path_to_gcwiggle_path(fasta_path):
    """e.g. ref.fasta -> ref.gc50.wig.gz"""
    return re.sub('\.[^.]+$', f'.gc{window_size}.wig.gz', fasta_path)


def seqzutils_gc_wiggle(fasta_path, outfile_path=None, window_size=50):
    if outfile_path is None:
        outfile_path = fasta_path_to_gcwiggle_path(fasta_path)

    run_sequenza_utils(['gc_wiggle', 
                        '-f', fasta_path, 
                        '-o', outfile_path, 
                        '-w', str(window_size)])


def bam2seqz_sanity_check(force_outfile_name, outfile_path, refver, 
                          fasta_path, gcwiggle_path):
    if not force_outfile_name:
        if not outfile_path.endswith('.seqz.gz'):
            raise Exception(f'Output file name must end with ".seqz.gz".')

    if not ((
                (refver is not None) and 
                (fasta_path is None) and 
                (gcwiggle_path is None)) or
            (
                (refver is None) and
                (fasta_path is not None) and
                (gcwiggle_path is not None))):
        raise Exception(
            f'Argument usage must be one of following:\n'
            f'1) "refver" set, "fasta_path" and "gcwiggle_path" not set;\n'
            f'2) "refver" not set, "fasta_path" and "gcwiggle_path" set.'
            )


@common.get_deco_num_set(('refver', 'fasta_path'), 1)
def seqzutils_bam2seqz(tbam_path, nbam_path, outfile_path, 
                       refver=None, fasta_path=None, gcwiggle_path=None, 
                       force_outfile_name=False,
                       chrom=None, start0=None, end0=None,
                       hom=DEFAULT_HOM, het=DEFAULT_HET):
    def seqz_filter(infile_path, outfile_path, chrom, start0, end0):
        with \
                open(infile_path, 'rt') as infile, \
                common.openfile(outfile_path, 'w') as outfile:
            header_line = next(infile)
            outfile.write(header_line)

            for line in infile:
                linesp = line.split('\t')
                chrom_line = linesp[0]
                pos0_line = int(linesp[1]) - 1
                if region_check(chrom, start0, end0, chrom_line, pos0_line):
                    outfile.write(line)

    def region_check(chrom, start0, end0, chrom_line, pos0_line):
        if chrom is not None:
            if start0 is None or end0 is None:
                return (chrom == chrom_line)
            else:
                return (chrom_line == chrom and 
                        pos0_line >= start0 and
                        pos0_line < end0)
        else:
            return True

    def run_bam2seqz(outfile_path, tbam_path, nbam_path, gcwiggle_path, 
                     fasta_path, hom, het, region):
        args = ['bam2seqz',
                '-t', tbam_path,
                '-n', nbam_path,
                '-gc', gcwiggle_path,
                '-F', fasta_path,
                '-o', outfile_path,
                '--hom', str(hom),
                '--het', str(het),
                '--samtools', common.SAMTOOLS,
                '--tabix', common.TABIX]
        if region is not None:
            args.extend(['-C', region])

        run_sequenza_utils(args)

    def get_fasta_gcwiggle_paths(refver, fasta_path, gcwiggle_path):
        if refver is not None:
            fasta_path = common.DEFAULT_FASTA_PATHS[refver]
            gcwiggle_path = DEFAULT_GCWIG_PATHS[refver]
        else:
            # fasta_path is not None in this case
            if gcwiggle_path is None:
                gcwiggle_path = fasta_path_to_gcwiggle_path(fasta_path)

        return fasta_path, gcwiggle_path

    def get_region(chrom, start0, end0):
        if chrom is not None:
            if (start0 is not None) and (end0 is not None):
                region = f'{chrom}:{start0 + 1}-{end0}'
            else:
                region = chrom
        else:
            region = None

        return region

    # sanity check
    bam2seqz_sanity_check(force_outfile_name, outfile_path, refver, 
                          fasta_path, gcwiggle_path)

    # parameter refinements
    fasta_path, gcwiggle_path = get_fasta_gcwiggle_paths(refver, fasta_path, 
                                                         gcwiggle_path)
    region = get_region(chrom, start0, end0)

    # main
    if region is None:
        run_bam2seqz(outfile_path, tbam_path, nbam_path, gcwiggle_path, 
                     fasta_path, hom, het, region)
    else:
        with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmpdir:
            fifo_path = os.path.join(tmpdir, 'fifo')
            os.mkfifo(fifo_path)

            p_fiforeader = multiprocessing.Process(
                target=seqz_filter,
                args=(fifo_path, outfile_path, chrom, start0, end0))
            p_fifowriter = multiprocessing.Process(
                target=run_bam2seqz,
                args=(fifo_path, tbam_path, nbam_path, gcwiggle_path, 
                      fasta_path, hom, het, region))

            p_fiforeader.start()
            p_fifowriter.start()

            p_fifowriter.join()
            p_fiforeader.join()




import sys
import re
import itertools
import collections

import pysam
import numpy as np
import pandas as pd

import handygenome.tools as tools
import handygenome.align.alignhandler as alignhandler


CIGAROPDICT = {
    "M": 0,
    "I": 1,
    "D": 2,
    "N": 3,
    "S": 4,
    "H": 5,
    "P": 6,
    "=": 7,
    "X": 8,
    "B": 9,
}
CIGARPAT = re.compile(f'([0-9]+)([{"".join(CIGAROPDICT.keys())}])')
CIGAR_WALK_DICT = { # (target, query)
    0: (True, True),
    1: (False, True),
    2: (True, False),
    3: (True, False),
    4: (False, True),
    5: (False, False),
    6: (False, False),
    7: (True, True),
    8: (True, True),
}
CIGAROPS_TARGETONLY = {2, 3}
CIGAROPS_QUERYONLY = {1, 4}


class NoMDTagError(Exception):
    pass


###################################

def get_read_sortkey(chromdict):
    def sortkey(read):
        return tools.coord_sortkey(read.reference_name, read.reference_start, chromdict)

    return sortkey


ReadUID = collections.namedtuple(
    'ReadUID',
    ('qname', 'flag', 'chrom', 'pos0'),
)


def get_uid(read):
    return ReadUID(read.query_name, read.flag, read.reference_name, read.reference_start)
    #return (read.query_name, read.flag, read.reference_name, read.reference_start)


def get_read_dict(read_iter):
    return {get_uid(x): x for x in read_iter}


def readfilter_bad_read(read):
    """returns True if the read is to be included"""

    return not (
        (not read.is_paired)
        or read.is_unmapped
        or read.mate_is_unmapped
        or read.is_duplicate
    )


def readfilter_pileup(read):
    return (
        readfilter_bad_read(read) and 
        (not read.is_supplementary) and 
        (not read.is_secondary)
    )


def get_chromlen_from_bam(bam, chrom):
    return bam.lengths[bam.references.index(chrom)]


def get_fetch(bam, chrom, start, end, readfilter=None):
    """
    - returns a generator
    - similar to bam.fetch iterator except read filtering
    """
    # set default filtering function
    if readfilter is None:
        readfilter = readfilter_bad_read
    # modify start and end
    start = max(0, start)
    end = min(get_chromlen_from_bam(bam, chrom), end)
    # main
    for read in bam.fetch(chrom, start, end):
        if readfilter(read):
            yield read


# cigar-related classes and functions

#class CigarUnit:
#    def __init__(self, opcode, opstring, count):
#        self.opcode = opcode
#        self.opstring = opstring
#        self.count = count

class Cigar:
    CIGAROPDICT_ITEMS = [
        ("M", 0),
        ("I", 1),
        ("D", 2),
        ("N", 3),
        ("S", 4),
        ("H", 5),
        ("P", 6),
        ("=", 7),
        ("X", 8),
        ("B", 9),
    ]
    CIGAROPDICT = dict(CIGAROPDICT_ITEMS)
    CIGAROPDICT_REV = dict((x[1], x[0]) for x in CIGAROPDICT_ITEMS)
    CIGARPAT = re.compile(f'([0-9]+)([{"".join(CIGAROPDICT.keys())}])')
    CIGAROPS_TARGETONLY = {2, 3}
    CIGAROPS_QUERYONLY = {1, 4}

    # constructors #
    @classmethod
    def from_cigarstring(cls, cigarstring):
        result = cls()
        result.cigartuples = cls.cigarstring_to_cigartuples(cigarstring)
        return result
    
    @classmethod
    def from_cigartuples(cls, cigartuples):
        result = cls()
        result.cigartuples = cigartuples
        return result
    ################

    # converters
    @classmethod
    def cigarstring_to_cigartuples(cls, cigarstring):
        result = list()
        for (count, opstring) in cls.CIGARPAT.findall(cigarstring):
            opcode = cls.CIGAROPDICT[opstring]
            count = int(count)
            result.append((opcode, count))
        return result

    @classmethod
    def cigartuples_to_cigarstring(cls, cigartuples):
        buffer = list()
        for opcode, count in cigartuples:
            opstring = cls.CIGAROPDICT_REV[opcode]
            buffer.append(str(count))
            buffer.append(opstring)
        return "".join(buffer)

    # instance methods
    @property
    def cigarstring(self):
        return self.cigartuples_to_cigarstring(self.cigartuples)

    def iter_trailing_queryonly(self):
        for opcode, count in reversed(self.cigartuples):
            if opcode in CIGAROPS_QUERYONLY:
                yield opcode, count
            else:
                break

    def iter_leading_queryonly(self):
        for opcode, count in self.cigartuples:
            if opcode in CIGAROPS_QUERYONLY:
                yield opcode, count
            else:
                break

    def check_DN_outside(self):
        # cigar I may be outside!
        opstrings_rmclip = [x for x in self.opstrings if x not in "SH"]

        return opstrings_rmclip[0] in "DN" or opstrings_rmclip[-1] in "DN"

    def check_clip_inside(self):
        if len(self.opstrings) <= 2:
            return False
        else:
            return not {"S", "H"}.isdisjoint(self.opstrings[1:-1])

    def check_unexpected_pattern(self):
        return (
            self.check_DN_outside()
            or self.check_clip_inside()
            or (not {"B", "P"}.isdisjoint(self.opstrings))
        )


#def get_cigartuples(cigarstring):
#    return [
#        (CIGAROPDICT[cigarop], int(count))
#        for (count, cigarop) in CIGARPAT.findall(cigarstring)
#    ]


#CigarWalk = collections.namedtuple(
#    'CigarWalk', 
#    ('cigartuple', 'target_range0', 'query_range0'),
#)
#def walk_cigar(cigartuples, target_start0):
#    query_start0 = 0
#
#    target_end0 = target_start0
#    query_end0 = query_start0
#
#    for cigartup in cigartuples:
#        opcode, count = cigartup
#        walk_target, walk_query = CIGAR_WALK_DICT[opcode]
#        if walk_target:
#            target_end0 += count
#        if walk_query:
#            query_end0 += count
#
#        target_range0 = range(target_start0, target_end0)
#        query_range0 = range(query_start0, query_end0)
#
#        target_start0 = target_end0
#        query_start0 = query_end0
#
#        yield CigarWalk(cigartup, target_range0, query_range0)
#
#
#def split_cigar(cigartuples, reference_start0, split_range0):
#    def add_buffer(tuple_list, queryonly_buffer):
#        while queryonly_buffer:
#            tuple_list.append(queryonly_buffer.pop(0).cigartuple)
#
#    def add_current(tuple_list, cigarwalk):
#        tuple_list.append(cigarwalk.cigartuple)
#
#    tuples_before = list()
#    tuples_within = list()
#    tuples_after = list()
#    reference_pointer = reference_start0
#    queryonly_buffer = list()
#
#    for cigarwalk in walk_cigar(cigartuples, reference_start0):
#        reference_pointer += len(cigarwalk.target_range0)
#
#        if len(cigarwalk.target_range0) == 0:
#            queryonly_buffer.append(cigarwalk)
#        else:
#            # current cigarwalk entirely before split_range0
#            if cigarwalk.target_range0.stop <= split_range0.start:
#                add_buffer(tuples_before, queryonly_buffer)
#                add_current(tuples_before, cigarwalk)
#            # current cigarwalk entirely after split_range0
#            elif cigarwalk.target_range0.start >= split_range0.stop:
#                add_buffer(tuples_after, queryonly_buffer)
#                add_current(tuples_after, cigarwalk)
#            # current cigarwalk overlaps split_range0
#            else:
#                # handle buffer
#                if cigarwalk.target_range0.start < split_range0.start:
#                    add_buffer(tuples_before, queryonly_buffer)
#                else:
#                    add_buffer(tuples_within, queryonly_buffer)
#                # add to tuples_before
#                if cigarwalk.target_range0.start < split_range0.start:
#                    tuple_to_add = (
#                        cigarwalk.cigartuple[0], 
#                        split_range0.start - cigarwalk.target_range0.start,
#                    )
#                    tuples_before.append(tuple_to_add)
#                # add to tuples_after
#                if cigarwalk.target_range0.stop > split_range0.stop:
#                    tuple_to_add = (
#                        cigarwalk.cigartuple[0], 
#                        cigarwalk.target_range0.stop - split_range0.stop,
#                    )
#                    tuples_after.append(tuple_to_add)
#                # add to tuples_within
#                tuple_to_add = (
#                    cigarwalk.cigartuple[0], 
#                    (
#                        min(cigarwalk.target_range0.stop, split_range0.stop)
#                        - max(cigarwalk.target_range0.start, split_range0.start)
#                    )
#                )
#                tuples_within.append(tuple_to_add)
#                    
#    # now reference_pointer is equal to read.reference_end
#    if queryonly_buffer:
#        if reference_pointer < split_range0.start:
#            add_buffer(tuples_before, queryonly_buffer)
#        else:
#            if reference_pointer < split_range0.stop:
#                add_buffer(tuples_within, queryonly_buffer)
#            else:
#                add_buffer(tuples_after, queryonly_buffer)
#
#    return tuples_before, tuples_within, tuples_after


#def reverse_range(rng):
#    if rng.step == 1:
#        start = rng.stop - 1
#        stop = rng.start - 1
#        return range(start, stop, -1)
#    elif rng.step == -1:
#        start = rng.stop + 1
#        stop = rng.start + 1
#        return range(start, stop, 1)
#
#
##def check_overlaps_forward_nonzero(rng1, rng2):
##    return (
##        (rng1.start < rng2.stop)
##        and (rng1.stop > rng2.start)
##    )
#
#
#def check_overlaps_forward_nonzero(rng1, rng2):
#    return common.check_overlaps(rng1.start, rng1.stop, rng2.start, rng2.stop)
#
#
#def check_overlaps(rng1, rng2):
#    """
#    Meaning of range
#        - Assumes range.step is either 1 or -1
#        - "current position" travels, beginning from "range.start" toward "range.stop".
#        - "cursor" position
#            - When range.step == 1: immediately left to "current position"
#            - When range.step == -1: immediately right to "current position"
#        - Area swept by "cursor" is the area represented by range object.
#        - When len(range) == 0:
#            - When range.step == 1: 0-length interval between (range.start - 1) and (range.start)
#            - When range.step == -1: 0-length interval between (range.start) and (range.start + 1)
#
#    Meaning of overlap with 0-length range
#        - consider as overlap when the interspace is contained between ALIGNED read bases (not insertion)
#
#    Args:
#        rng1, rng2: python range object.
#    """
#    # convert into positive-step range
#    if rng1.step == -1:
#        rng1 = reverse_range(rng1)
#    if rng2.step == -1:
#        rng2 = reverse_range(rng2)
#
#    # main
#    if len(rng1) == 0 and len(rng2) == 0:
#        return rng1.start == rng2.start
#    else:
#        return check_overlaps_forward_nonzero(rng1, rng2)
#
#
#def check_spans(query_range0, target_range0):
#    """Checks if query spans target"""
#
#    if len(query_range0) == 0:
#        return False
#    else:
#        if len(target_range0) == 0:
#            return check_overlaps(query_range0, target_range0)
#        else:
#            return (
#                min(query_range0) <= min(target_range0) and
#                max(query_range0) >= max(target_range0)
#            )


def get_aligned_pairs(pos, cigartuples):
    aligned_pairs = list()
    querypos0 = 0
    refpos0 = pos - 1

    for (cigarop, count) in cigartuples:
        if cigarop in (2, 3):  # D, N
            for _ in range(count):
                aligned_pairs.append((None, refpos0))
                refpos0 += 1
        elif cigarop in (1, 4, 5):  # I, S, H
            for _ in range(count):
                aligned_pairs.append((querypos0, None))
                querypos0 += 1
        elif cigarop in (0, 7, 8):  # M, =, X
            for _ in range(count):
                aligned_pairs.append((querypos0, refpos0))
                querypos0 += 1
                refpos0 += 1

    return aligned_pairs


def get_pairs_dict(read, fasta=None, skip_refseq=False, set_cigarop=False, donot_use_existing_MD=False):
    """Value of read.get_aligned_pairs method:
        cigar N: represented as (None, None)
        cigar H: not represented (len(read.get_aligned_pairs()) == 37 for 114H37M)

    Returns:
        Dict {
            'querypos0': A tuple of query positions
            'refpos0': A tuple of reference genome positions
            'refseq': A tuple of reference sequences (lowercase with mismatch)
            'cigarop': A tuple of cigar operations, as integers
        }
    """

    if (
        donot_use_existing_MD or
        (not read.has_tag("MD"))
    ):
        raw = read.get_aligned_pairs(with_seq=False)
        zipped = zip(*raw)
        pairs_dict = dict()
        pairs_dict["querypos0"] = next(zipped)
        pairs_dict["refpos0"] = next(zipped)
        if not skip_refseq:
            pairs_dict["refseq"] = get_pairs_dict_refseq(read, fasta, raw)
    else:
        raw = read.get_aligned_pairs(with_seq=(not skip_refseq))
        zipped = zip(*raw)
        pairs_dict = dict()
        pairs_dict["querypos0"] = next(zipped)
        pairs_dict["refpos0"] = next(zipped)
        if not skip_refseq:
            pairs_dict["refseq"] = next(zipped)

    # cigarop
    if set_cigarop:
        pairs_dict["cigarop"] = tuple(
            itertools.chain.from_iterable(
                itertools.repeat(cigarop, count)
                for (cigarop, count) in read.cigartuples
                if cigarop != 5
            )
        )

        if len(pairs_dict["cigarop"]) != len(pairs_dict["refpos0"]):
            raise Exception(
                f"The length of cigar operations differ from that of "
                f"aligned pairs:\n"
                f"{read.to_string()}"
            )

    return pairs_dict


def get_pairs_dict_refseq(read, fasta, raw_aligned_pairs):
    refseq = list()

    non_None_refpos0 = [
        refpos0 for (querypos0, refpos0) in raw_aligned_pairs if refpos0 is not None
    ]
    if len(non_None_refpos0) == 0:
        fetched_seq = ""
    else:
        fetched_seq = fasta.fetch(
            read.reference_name, non_None_refpos0[0], non_None_refpos0[-1] + 1
        )
    it = iter(fetched_seq)
    for querypos0, refpos0 in raw_aligned_pairs:
        if refpos0 is None:
            refseq.append(None)
        else:
            if querypos0 is None:
                read_base = None
            else:
                read_base = read.query_sequence[querypos0].upper()
            ref_base = next(it).upper()

            if read_base is None:
                refseq.append(ref_base)
            else:
                if read_base == ref_base:
                    refseq.append(ref_base)
                else:
                    refseq.append(ref_base.lower())

    return refseq


# def get_MM(read, pairs_dict):
#    if pairs_dict is None:
#        MM = None
#    else:
#        cigarM_indices = [
#            True
#            if (tup[0] is not None and tup[1] is not None) else
#            False
#            for tup in zip(pairs_dict['querypos0'], pairs_dict['refpos0'])]
#        pairs_dict_seq_subset = itertools.compress(
#            pairs_dict['refseq'], cigarM_indices)
#        MM = len([x for x in pairs_dict_seq_subset if x.islower()])
#
#    return MM


#####################################################


def check_cigar_clip_inside(cigartuples):
    if len(cigartuples) <= 2:
        return False
    else:
        cigarops_inside = [x[0] for x in cigartuples[1:-1]]
        return not {4, 5}.isdisjoint(cigarops_inside)


def check_cigar_DN_outside(cigarops):
    # cigar I can be outside!
    cigarops_rmclip = [x for x in cigarops if x not in (4, 5)]
    return cigarops_rmclip[0] in (2, 3) or cigarops_rmclip[-1] in (2, 3)


def check_unexpected_cigar_pattern(cigartuples):
    cigarops = [x[0] for x in cigartuples]
    if check_cigar_DN_outside(cigarops):
        return True
    elif check_cigar_clip_inside(cigartuples):
        return True
    elif "B" in cigarops:
        return True
    elif "P" in cigarops:
        return True
    else:
        return False


#############################################################


def get_padded_seqs(read, fasta):
    if check_unexpected_cigar_pattern(read.cigartuples):
        raise Exception(
            f"Unexpected cigar pattern with the input read: " f"{read.to_string()}"
        )

    ref_seq_padded = list()
    read_seq_padded = list()

    ref_seq = list(
        fasta.fetch(read.reference_name, read.reference_start, read.reference_end)
    )
    read_seq = list(read.query_sequence)

    for idx_cigar, tup in enumerate(read.cigartuples):
        op, l = tup  # op: operation ; l: length
        if op in (0, 7, 8):  # cigar M, =, X
            for i in range(l):
                ref_seq_padded.append(ref_seq.pop(0))
                read_seq_padded.append(read_seq.pop(0))
        elif op == 1:  # cigar I
            ref_seq_padded.append(None)
            ins_seq_buffer = list()
            for i in range(l):
                ins_seq_buffer.append(read_seq.pop(0))
            read_seq_padded.append("".join(ins_seq_buffer))
        elif op == 2:  # cigar D
            read_seq_padded.append(None)
            del_seq_buffer = list()
            for i in range(l):
                del_seq_buffer.append(ref_seq.pop(0))
            ref_seq_padded.append("".join(del_seq_buffer))
        elif op == 3:  # cigar N
            del ref_seq[:l]
        elif op == 4:  # cigar S
            del read_seq[:l]
        elif op == 5:  # cigar H
            pass

    return ref_seq_padded, read_seq_padded


def get_MD(read, fasta=None, ref_seq_padded=None, read_seq_padded=None):
    if (ref_seq_padded is None) or (read_seq_padded is None):
        ref_seq_padded, read_seq_padded = get_padded_seqs(read, fasta)

    MD_list = list()
    identical = 0
    for ref_base, read_base in zip(ref_seq_padded, read_seq_padded):
        if ref_base is not None and read_base is not None:
            if read_base == "=":  # means a base identical to reference base
                identical += 1
            else:
                if ref_base == read_base:  # identical match
                    identical += 1
                else:  # different match
                    MD_list.append(str(identical))
                    identical = 0
                    MD_list.append(ref_base)
        elif ref_base is None:  # cigar I
            pass
        elif read_base is None:  # cigar D
            MD_list.append(str(identical))
            identical = 0
            MD_list.append("^" + ref_base)

    MD_list.append(str(identical))

    return "".join(MD_list)


def get_NM(read, fasta=None, ref_seq_padded=None, read_seq_padded=None):
    if (ref_seq_padded is None) or (read_seq_padded is None):
        ref_seq_padded, read_seq_padded = get_padded_seqs(read, fasta)

    NM = 0
    for ref_base, read_base in zip(ref_seq_padded, read_seq_padded):
        if ref_base is None:  # cigar I
            NM += len(read_base)
        elif read_base is None:  # cigar D
            NM += len(ref_base)
        else:
            if read_base == "=":  # means a base identical to reference base
                pass
            else:
                if ref_base == read_base:  # identical match
                    pass
                else:  # different match
                    NM += 1

    return NM


def get_cigarN_pairs_indexes(read):
    non_cigarN_pairs_indexes = list()
    cigarN_pairs_indexes = list()
    end0 = 0
    for key, subiter in itertools.groupby(read.cigartuples, key=(lambda x: x[0] == 3)):
        if key:  # cigar N
            start0 = end0
            for cigarop, cigarlen in subiter:
                end0 += cigarlen
            cigarN_pairs_indexes.append(slice(start0, end0))
        else:
            start0 = end0
            for cigarop, cigarlen in subiter:
                if cigarop not in (5, 6):  # cigar H, P
                    end0 += cigarlen
            non_cigarN_pairs_indexes.append(slice(start0, end0))

    return non_cigarN_pairs_indexes, cigarN_pairs_indexes


def get_MD_from_pairs_dict(read, pairs_dict, non_cigarN_pairs_indexes=None):
    # remove cigar N from pairs_dict
    if 'N' in read.cigarstring:
        if non_cigarN_pairs_indexes is None:
            non_cigarN_pairs_indexes = get_cigarN_pairs_indexes(read)[0]

        pairs_dict = {
            key: list(
                itertools.chain.from_iterable(
                    val[sl] for sl in non_cigarN_pairs_indexes
                )
            )
            for key, val in pairs_dict.items()
        }

    # main
    MD_list = list()
    identical = 0

    for key, subiter in itertools.groupby(
        zip(
            pairs_dict['refpos0'],
            pairs_dict['querypos0'],
            pairs_dict['refseq'],
        ),
        key=(lambda x: ((x[0] is None), (x[1] is None))),
    ):
        if (not key[0]) and key[1]:  # D
            MD_list.append(str(identical))
            identical = 0
            MD_list.append('^' + ''.join(x[2] for x in subiter))
        elif (not key[0]) and (not key[1]):  # match
            for refpos0, querypos0, refseq in subiter:
                if refseq.islower():
                    MD_list.append(str(identical))
                    identical = 0
                    MD_list.append(refseq.upper())
                else:
                    identical += 1
        elif key[0] and (not key[1]):  # I, S
            continue
    
    MD_list.append(str(identical))

    return "".join(MD_list)


def get_NM_from_pairs_dict(read, pairs_dict):
    NM = 0
    for refpos0, querypos0, refseq in zip(
        pairs_dict['refpos0'], 
        pairs_dict['querypos0'], 
        pairs_dict['refseq'],
    ):
        ref_missing = (refpos0 is None)
        query_missing = (querypos0 is None)
        if ref_missing != query_missing:
            NM += 1
        else:
            if refseq.islower():
                NM += 1
                  
    cigar_stats = read.get_cigar_stats()
    NM -= cigar_stats[0][3]  # N
    NM -= cigar_stats[0][4]  # S

    return NM


def set_NMMD(read, fasta=None, pairs_dict=None, non_cigarN_pairs_indexes=None):
    if pairs_dict is None:
        pairs_dict = get_pairs_dict(
            read, fasta=fasta, skip_refseq=False, set_cigarop=False, donot_use_existing_MD=True,
        )

    NM = get_NM_from_pairs_dict(read, pairs_dict)
    MD = get_MD_from_pairs_dict(read, pairs_dict, non_cigarN_pairs_indexes=non_cigarN_pairs_indexes)

    read.set_tag(tag='NM', value_type='i', value=NM)
    read.set_tag(tag='MD', value_type='Z', value=MD)


def set_NMMD_old(read, fasta):
    ref_seq_padded, read_seq_padded = get_padded_seqs(read, fasta)
    NM = get_NM(read, ref_seq_padded=ref_seq_padded, read_seq_padded=read_seq_padded)
    MD = get_MD(read, ref_seq_padded=ref_seq_padded, read_seq_padded=read_seq_padded)
    read.set_tag(tag='NM', value_type='i', value=NM)
    read.set_tag(tag='MD', value_type='Z', value=MD)


def get_NMMD(read, fasta):
    ref_seq_padded, read_seq_padded = get_padded_seqs(read, fasta)
    NM = get_NM(read, ref_seq_padded=ref_seq_padded, read_seq_padded=read_seq_padded)
    MD = get_MD(read, ref_seq_padded=ref_seq_padded, read_seq_padded=read_seq_padded)
    return NM, MD


def get_pairorient_substring(read, mate=False):
    if mate:
        orientation = "r" if read.mate_is_reverse else "f"
        read12 = "2" if read.is_read1 else "1"
    else:
        orientation = "r" if read.is_reverse else "f"
        read12 = "1" if read.is_read1 else "2"

    return orientation + read12


def get_pairorient(read):
    if read.mate_is_unmapped:
        pairorient = None
    else:
        if read.template_length == 0:  # This occurs when mate chroms differ
            pairorient = None
        else:
            if read.template_length > 0:
                substring1 = get_pairorient_substring(read, mate=False)
                substring2 = get_pairorient_substring(read, mate=True)
            elif read.template_length < 0:
                substring1 = get_pairorient_substring(read, mate=True)
                substring2 = get_pairorient_substring(read, mate=False)

            pairorient = substring1 + substring2

    return pairorient


#####


def check_TRA(read):
    return read.reference_name != read.next_reference_name


def check_primary_alignment(read):
    return (not read.is_secondary) and (not read.is_supplementary)


#####


def get_fiveprime_end(read):
    if read.is_reverse:
        fiveprime_end = read.reference_end - 1
    else:
        fiveprime_end = read.reference_start

    return fiveprime_end


def get_threeprime_end(read):
    if read.is_reverse:
        threeprime_end = read.reference_start
    else:
        threeprime_end = read.reference_end - 1

    return threeprime_end


#####


def get_template_range0(read, with_softclip=False):
    """Returns a range object representing the interval between 5' ends of
        the read pairs.
    It is in 0-based, half-open format (bed format).
    """

    if read.template_length == 0:
        """
        TLEN == 0 cases include
            - different chromosomes between read pairs
            - two reads are on different strands and have identical 5' end
                base position
        """
        template_range0 = None
    else:
        if read.template_length > 0:
            start = get_fiveprime_end(read)
            end = start + read.template_length
        elif read.template_length < 0:
            end = get_fiveprime_end(read) + 1  # bed format
            start = end + read.template_length

        template_range0 = range(start, end)

    return template_range0


def get_softclip_ends_range0(read):
    if read.cigartuples[0][0] == 4:  # first cigar is S
        start = read.reference_start - read.cigartuples[0][1]
    else:
        start = read.reference_start

    if read.cigartuples[-1][0] == 4:  # last cigar is S
        end = read.reference_end + read.cigartuples[-1][1]
    else:
        end = read.reference_end

    return range(start, end)


Clipspec = collections.namedtuple(
    "Clipspec", 
    #("pos0", "is_forward", "seq", "qual", "qname"),
    ("pos0", "is_forward", "seq", "qual", "readuid"),
)


def get_softclip_specs(read):
    readuid = get_uid(read)
    clipspec_list = list()

    if read.cigartuples[0][0] == 4:
        pos0 = read.reference_start
        cliplen = read.cigartuples[0][1]
        seq = read.query_sequence[:cliplen][::-1]
        qual = list(read.query_qualities)[:cliplen][::-1]
        is_forward = False
        clipspec_list.append(Clipspec(pos0, is_forward, seq, qual, readuid))

    if read.cigartuples[-1][0] == 4:
        #start1 = read.reference_end + 1
        pos0 = read.reference_end - 1
        cliplen = read.cigartuples[-1][1]
        seq = read.query_sequence[-cliplen:]
        qual = list(read.query_qualities)[-cliplen:]
        is_forward = True
        clipspec_list.append(Clipspec(pos0, is_forward, seq, qual, readuid))

    return clipspec_list


#####


def check_mateless(read):
    return (not read.is_paired) or read.mate_is_unmapped


def get_primary_mate_candidate(bam, read):
    primary_mate_candidate = list()
    for new_read in bam.fetch(
        read.next_reference_name,
        read.next_reference_start,
        read.next_reference_start + 1,
    ):
        if check_primary_alignment(new_read):
            if (new_read.query_name == read.query_name) and (
                read.compare(new_read) != 0
            ):
                # read.compare method returns 0 if the two reads are identical
                primary_mate_candidate.append(new_read)

    return primary_mate_candidate


def get_primary_mate(read, bam):
    """Returns:
        None if the input read is unpaired or the mate is unmapped.
        Else:
            1) Performs fetch with input bam in a position specified by RNEXT
                and PNEXT of the input read.
            2) From the fetch result, picks reads with the same QNAME as the
                input read.
            3) If there are only one picked read, returns it.
                Otherwise, raises an exception.
    """

    if check_mateless(read):
        mate = None
    else:
        primary_mate_candidate = get_primary_mate_candidate(bam, read)

        if len(primary_mate_candidate) == 0:
            raise Exception(
                f"No primary mate candidate found for this read:\n"
                f"{read.to_string()}"
            )
        elif len(primary_mate_candidate) == 1:
            mate = primary_mate_candidate[0]
        elif len(primary_mate_candidate) > 1:
            mate_candidate_strings = "\n".join(
                [read.to_string() for read in primary_mate_candidate]
            )
            raise Exception(
                f"More than one primary mate candidates found for "
                f"the input read:\n"
                f"{read.to_string()}\n"
                f"Detected mate candidates for this read:\n"
                f"{mate_candidate_strings}"
            )

    return mate


def check_cigarN_includes_range(read, start0, end0):
    cigarN_target_ranges = list()
    for cigarwalk in alignhandler.walk_cigar(
        read.cigartuples, read.reference_start, with_cigartup=True,
    ):
        if cigarwalk.cigartup[0] == 3:
            cigarN_target_ranges.append(cigarwalk.target)
                    
    return any(
        start0 >= rng.start and end0 <= rng.stop
        for rng in cigarN_target_ranges
    )


def check_all_queryonly(read):
    return set(x[0] for x in read.cigartuples).issubset(alignhandler.CIGAROPS_QUERYONLY)






import pysam
import Bio.Seq
import pandas as pd

import handygenome.signature.misc as signature_misc
import handygenome.deco as deco


#@deco.get_deco_arg_choices({'refver': common.AVAILABLE_REFVERS_PLUSNONE})
@deco.get_deco_num_set(('fasta', 'refver'), 1)
def get_sbs_context(
    chrom, pos, fasta=None, refver=None, pre=1, post=1, to_pyrimidine=False,
):
    if fasta is None:
        fasta = refgenome.get_fasta(refver)

    start = pos - 1 - pre
    end = pos + post
    seq = fasta.fetch(chrom, start, end).upper()
    center_base = seq[pre]
    if (center_base in 'AG') and to_pyrimidine:
        seq = Bio.Seq.reverse_complement(seq)

    return seq


def get_sbs96_catalogue_vcfspecs(vcfspec_iter, refver):
    """Multialleleic records and non-snv records are ignored."""

    def skip_vcfspec(vcfspec):
        """skipped if True"""
        if len(vcfspec.alts) != 1:  # multialleleic record
            return True
        else:
            if not (len(vcfspec.ref) == len(vcfspec.alts[0]) == 1):  
                # not a snv
                return True
            else:
                return False
                
    fasta = refgenome.get_fasta(refver)
    catalogue_keys = signature_misc.get_catalogue_keys('sbs96')
    data = dict((x, 0) for x in catalogue_keys)

    for vcfspec in vcfspec_iter:
        if skip_vcfspec(vcfspec):
            continue
        # get sequences
        context = get_sbs_context(vcfspec.chrom, vcfspec.pos, fasta=fasta, 
                                  pre=1, post=1, to_pyrimidine=False)
        alt = vcfspec.alts[0]
        # get reverse complements if REF is purine
        if vcfspec.ref in 'AG':
            context = Bio.Seq.reverse_complement(context)
            alt = Bio.Seq.reverse_complement(alt)
        # get catalogue key and update data
        cat_key = f'{context[0]}[{context[1]}>{alt}]{context[2]}'
        data[cat_key] += 1

    result = pd.Series(data=data, index=catalogue_keys)

    return result




import re
import os
import pickle
import itertools
import collections
import functools

import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
import scipy.stats
import scipy.optimize
import scipy.interpolate
import sklearn.mixture

import handygenome
import handygenome.tools as tools
import handygenome.deco as deco
import handygenome.logutils as logutils
#import handygenome.tools as tools
import handygenome.peakutils as peakutils
from handygenome.peakutils import HistPeaks, DensityPeaks, DensityGenerationFailure, NoPeakError
import handygenome.refgenome.refgenome as refgenome
import handygenome.variant.variantplus as libvp
from handygenome.variant.variantplus import VCFDataFrame
from handygenome.genomedf.genomedf import GenomeDataFrame 
from handygenome.cnv.segment import SegmentDataFrame


##############
# exceptions #
##############

class FailedNoTroughMethod(Exception):
    pass


##########
# basics #
##########

def get_baf_from_vaf(vafs):
    """Args:
        vafs: 
            - 1st dimension means genomic positions, 2nd dimension means allelic vafs
            - Desired format contains vaf values for all alleles. 
            (ndim == 2, shape[1] >= 2)
            - If values for only one allele is present, 
            (ndim == 1 or (ndim == 2 and shape[1] == 1)) 
            a transformed copy of input is generated which has values for 
            all alleles.
    """
    # preprocessing "vafs" into a 2d array
    vafs = np.atleast_1d(vafs)
    if vafs.ndim == 1:
        vafs = np.stack([vafs, 1 - vafs], axis=1)
    elif vafs.ndim == 2:
        if vafs.shape[1] == 1:
            vafs = np.stack([vafs[:, 0], 1 - vafs[:, 0]], axis=1)
    else:
        raise Exception(f'Input array must not have ndim larger than 2.')
        
    # main
    vafs = np.sort(vafs, axis=1)
    bafs = vafs[:, :-1]
    return bafs


def get_baf_indexes(ploidy):
    assert ploidy >= 2
    return [f'baf{x}' for x in range(ploidy - 1)]


###################
# BAFRawDataFrame #
###################

class BAFRawDataFrame(GenomeDataFrame):
    def remove_overlapping_rows_slow(self):
        new_df = (
            self.cluster().df
            .groupby('Cluster', sort=False)
            .first()
            .reset_index(drop=True)
        )
        return self.spawn(new_df)

    def remove_overlapping_rows(self):
        """Intended to remove overlapping variants (e.g. pos 1 A>G & pos 1 AC>GT)
        """
        filtered_subgdfs = list()
        for chrom, subgdf in self.group_bychrom().items():
            while True:
                overlap_selector = np.concatenate(
                    [[False], (subgdf.starts[1:] <= subgdf.ends[:-1])], 
                    axis=0,
                )
                if not overlap_selector.any():
                    break
                else:
                    subgdf = subgdf.loc[~overlap_selector, :]

            filtered_subgdfs.append(subgdf)

        new_gdf = GenomeDataFrame.concat(filtered_subgdfs)
        return self.spawn(new_gdf.df)

    @property
    def baf_columns(self):
        """baf0 comes first, the baf1, then baf2, ..."""
        result = [x for x in self.columns if re.fullmatch(r'baf[0-9]+', x)]
        result.sort(key=(lambda x: int(x[3:])))  # baf0, baf1, baf2, ...
        return result

    def get_baf_array(self):
        """baf0 is on the left"""
        return self.df.loc[:, self.baf_columns].to_numpy()

    def get_segment(self, baf_idx, *args, **kwargs):
        seg = super().get_segment(*args, baf_idx, **kwargs)
        return BAFSegmentDataFrame.from_frame(seg.df, refver=self.refver)


@deco.get_deco_atleast1d(['sampleids'])
def get_bafdfs_from_vcf(
    vcf_path, 
    sampleids, 
    n_allele=2,
    nproc=1,
    exclude_other=False,
    prop=None,
    vpfilter=None,
    verbose=True,
):
    vafdf = libvp.get_vafdf(
        vcf_path, 
        sampleids, 
        n_allele=n_allele,
        nproc=nproc,
        exclude_other=exclude_other,
        prop=prop,
        vpfilter=vpfilter,
        verbose=verbose,
    )
    return get_bafdfs_from_vafdf(vafdf)


def get_bafdfs_from_vafdf(vafdf):
    assert isinstance(vafdf, VCFDataFrame)

    vaf_formatkeys = [f'{x}_vaf' for x in vafdf.allele_columns]
    result = dict()
    for sid in vafdf.samples:
        vafs = vafdf.get_format(sid, vaf_formatkeys).to_numpy()
        bafs = get_baf_from_vaf(vafs)

        src_data = dict()
        src_data['chroms'] = vafdf.chromosomes
        src_data['start0s'] = vafdf.starts
        src_data['end0s'] = vafdf.ends

        ploidy = bafs.shape[1] + 1
        for col_idx, baf_idx in enumerate(get_baf_indexes(ploidy)):
        #for idx in range(bafs.shape[1]):
            #baf_colname = f'baf{idx}'
            src_data[baf_idx] = bafs[:, col_idx]

        result[sid] = BAFRawDataFrame.from_data(refver=vafdf.refver, **src_data)

    return result


#######################
# BAFSegmentDataFrame #
#######################

class BAFSegmentDataFrame(SegmentDataFrame):
    distinfo_suffix = '_distinfo'
    corrected_baf_suffix = '_corrected'

    #@property
    #def baf_colname(self):
    #    result = [x for x in self.columns if re.fullmatch(r'baf[0-9]+', x)]
    #    if len(result) != 1:
    #        raise Exception(f'Invalid column names: does not have a unique baf-index-named column.')
    #    return result[0]

    #################################
    # baf_index and colname getters #
    #################################

    def get_baf_index(self):
        baf_strings = set()
        for x in self.annot_cols:
            mat = re.match(r'baf[0-9]+', x)
            if mat is not None:
                baf_strings.add(mat.group(0))
        if len(baf_strings) != 1:
            raise Exception(f'Could not identify baf index from column names.')

        return baf_strings.pop()

    def get_distinfo_colname(self, key):
        assert key in (
            'ndata', 
            'center', 
            'width',
            'peak_index',
            'height',
            'left_ips',
            'right_ips',
        )
        prefix = self.get_baf_index() + self.__class__.distinfo_suffix
        return prefix + '_' + key

    def get_corrected_baf_colname(self):
        return self.get_baf_index() + self.__class__.corrected_baf_suffix

    ##################
    # column getters #
    ##################

    def get_dist_center(self):
        return self[self.get_distinfo_colname('center')]

    def get_dist_ndata(self):
        return self[self.get_distinfo_colname('ndata')]

    def get_corrected_baf(self):
        return self[self.get_corrected_baf_colname()]

    #############
    # modifiers #
    #############

    @deco.get_deco_atleast1d(['distinfo_keys'])
    def add_rawdata_info(
        self, 
        baf_rawdata_gdf, 
        baf_idx,
        distinfo_keys=['ndata', 'center', 'width', 'left_ips', 'right_ips'],
        distinfo_kwargs=dict(),
        merge_methods=['mean', 'std'],
        nproc=1,
    ):
        if distinfo_keys is not None:
            merge_methods = merge_methods + [bafdistinfo_aggfunc]

        joined_gdf = self.drop_annots().join(
            baf_rawdata_gdf,
            right_gdf_cols=baf_idx,
            how='left',
            merge=merge_methods,
            suffixes={
                'bafdistinfo_aggfunc': self.__class__.distinfo_suffix,
            },
            merge_kwargs=distinfo_kwargs,
            nproc=nproc,
        )

        joined_df = joined_gdf._df
        bafdistinfo_key = str(baf_idx) + self.__class__.distinfo_suffix
        bafdistinfo_list = joined_df[bafdistinfo_key]
        for key in distinfo_keys:
            joined_df[f'{bafdistinfo_key}_{key}'] = [
                (
                    x['hetalt'][key] 
                    if x['hetalt'] is not None else
                    np.nan
                )
                for x in bafdistinfo_list
            ]
        joined_df.drop(bafdistinfo_key, axis=1, inplace=True)

        self.assign_frame(joined_df)
        #return self.spawn(joined_df)

    def add_corrected_baf(self, cutoff=0.41):
        centers = self.get_dist_center()
        centers[centers > cutoff] = 0.5
        self[self.get_corrected_baf_colname()] = centers

    def merge_low_ndata_segments(self, cutoff=30, nproc=1):
        ndata_colname = self.get_distinfo_colname('ndata')
        return super().merge_low_ndata_segments(
            ndata_colname=ndata_colname, 
            cutoff=cutoff, 
            nproc=nproc,
        )

    def incoporate_low_ndata_segments(self, cutoff=30, nproc=1):
        ndata_colname = self.get_distinfo_colname('ndata')
        return super().incoporate_low_ndata_segments(
            ndata_colname=ndata_colname, 
            cutoff=cutoff,
            nproc=nproc,
        )

    def get_filter_colname(self):
        baf_index = self.get_baf_index()
        return f'{baf_index}_selected'

    def add_filter(self, width_cutoff=None, center_cutoff=(0.4, None)):
        width_colname = self.get_distinfo_colname('width')
        center_colname = self.get_distinfo_colname('center')

        width_values = self[width_colname]
        width_selector = self.filter_helper(width_values, width_cutoff)

        center_values = self[center_colname]
        center_selector = self.filter_helper(center_values, center_cutoff)

        selector = np.logical_and(width_selector, center_selector)
        baf_index = self.get_baf_index()
        self[self.get_filter_colname()] = selector

    def get_filter(self):
        return self[self.get_filter_colname()]


#############################################################
# baf distribution analyzer - invalid region identification #
#############################################################

def get_bafdistinfo_maxpeakidx_wotrough(peakresult):
    return np.argmax(peakresult['peak_properties']['peak_heights'])


def get_bafdistinfo_maxpeakidx(peakresult, trough_x):
    relevant_selector = peakresult['peak_xs'] > trough_x
    relevant_peakheights = peakresult['peak_properties']['peak_heights'][relevant_selector]
    max_peakheight = np.max(relevant_peakheights)
    max_peak_idxs = np.nonzero(
        peakresult['peak_properties']['peak_heights'] == max_peakheight
    )[0]
    if len(max_peak_idxs) != 1:
        raise Exception(f'More than one maximal peaks right to the trough point')
    return max_peak_idxs[0]


def get_bafdistinfo_findtrough_old(peakresult):
    return peakresult['xs'][np.argmin(peakresult['ys'])]


def get_bafdistinfo_findtrough(peakresult):
    negative_ys = np.max(peakresult['ys']) - peakresult['ys']
    negative_peaks, negative_peak_properties = scipy.signal.find_peaks(negative_ys)
    if len(negative_peaks) == 0:
        return None
    else:
        return peakresult['xs'][np.min(negative_peaks)]

#BAFPEAKINFO_KEYS = (
#    'peak_index',
#    'center',
#    'height',
#    'left_ips',
#    'right_ips',
#    'ndata',
#)

#class BAFPeakInfo(collections.namedtuple('BAFPeakInfo', BAFPEAKINFO_KEYS)):
#    pass


def get_bafdistinfo_make_peakinfo(peakresult, peak_array_index, reltoheight=None):
    peakinfo_src = dict()

    peakinfo_src['peak_index'] = peak_array_index
    peakinfo_src['center'] = peakresult['peak_xs'][peak_array_index]
    peakinfo_src['height'] = peakresult['peak_ys'][peak_array_index]

    if reltoheight is None:
        left_key = 'left_ips'
        right_key = 'right_ips'
    else:
        left_key = 'left_ips_reltoheight'
        right_key = 'right_ips_reltoheight'

    peakinfo_src['left_ips'] = peakresult['xcoord_values'][left_key][peak_array_index]
    peakinfo_src['right_ips'] = peakresult['xcoord_values'][right_key][peak_array_index]
    peakinfo_src['width'] = peakinfo_src['right_ips'] - peakinfo_src['left_ips']

    peakinfo_src['ndata'] = peakresult['ndata'][peak_array_index]

    #return BAFPeakInfo(**peakinfo_src)
    return peakinfo_src


def check_homalt(peakinfo):
    return (peakinfo['left_ips'] < 0)


def make_result_new(peakresult, reltoheight):
    peakinfo_list = [
        get_bafdistinfo_make_peakinfo(peakresult, idx, reltoheight)
        for idx in range(peakresult.npeaks)
    ]
    homalt_selector = np.fromiter(map(check_homalt, peakinfo_list), dtype=bool)
    homalt_indexes = np.nonzero(homalt_selector)[0]
    hetalt_indexes = np.nonzero(~homalt_selector)[0]

    # homalt
    if len(homalt_indexes) == 0:
        #result['homalt'] = None
        homalt_peakinfo = None
    else:
        homalt_peakinfo = peakinfo_list[homalt_indexes[0]]

    # hetalt
    if len(hetalt_indexes) == 0:
        hetalt_peakinfo = None
    else:
        hetalt_candidates = [peakinfo_list[x] for x in hetalt_indexes]
        maxheight_arg = np.argmax([x['height'] for x in hetalt_candidates])
        hetalt_peakinfo = hetalt_candidates[maxheight_arg]

    return {'homalt': homalt_peakinfo, 'hetalt': hetalt_peakinfo}


def get_bafdistinfo(
    bafs, 

    decal=False,

    bw_method=None,
    xs=np.linspace(-0.05, 0.55, 121, endpoint=True),
    bins=np.linspace(-0.05, 0.55, 31, endpoint=True),
    reltoheight=None, 

    hist=False,
    return_peakresult=False,

    **kwargs,
):
    if decal:
        bafs = np.concatenate([bafs, 1 - bafs])

    if hist:
        peakresult = peakutils.find_hist_peaks(
            bafs, 
            reltoheight=reltoheight, 
            bins=bins,
            **kwargs,
        )
    else:
        peakresult = peakutils.find_density_peaks(
            bafs, 
            reltoheight=reltoheight, 
            xs=xs,
            bw_method=bw_method,
            **kwargs,
        )

    if decal:
        peakresult.undecal(0.5)
    peakresult = peakresult.remove_contained_peaks()
    bafdistinfo = make_result_new(peakresult, reltoheight)

    if return_peakresult:
        return bafdistinfo, peakresult
    else:
        return bafdistinfo


def bafdistinfo_aggfunc(
    bafs, 
    #keys=['ndata', 'center'],
    **kwargs,
):
    #assert set(keys).issubset(BAFPEAKINFO_KEYS)

    #assert bafs.ndim == 2
    #assert bafs.shape[1] == 1
    #bafs = np.asarray(bafs)[:, 0]
    bafs = np.atleast_1d(np.squeeze(np.asarray(bafs)))

    try:
        bafdistinfo = get_bafdistinfo(
            bafs, hist=False, return_peakresult=False, **kwargs
        )
    except (DensityGenerationFailure, NoPeakError):
        try:
            bafdistinfo = get_bafdistinfo(
                bafs, hist=True, return_peakresult=False, **kwargs
            )
        except NoPeakError as exc:
            raise Exception(f'Even histogram could not detect a peak from the baf data.') from exc

    return bafdistinfo


def draw_bafarray_breadth(
    bafs,

    distinfo_kwargs=dict(), 
    draw_peak_kwargs=dict(),

    hist=False,

    return_data=False,
):
    # prepare peakresult
    hist_bafdistinfo, hist_peakresult = get_bafdistinfo(
        bafs, 
        hist=True,
        return_peakresult=True,
        **distinfo_kwargs,
    )
    try:
        density_bafdistinfo, density_peakresult = get_bafdistinfo(
            bafs, 
            hist=False,
            return_peakresult=True,
            **distinfo_kwargs,
        )
    except DensityGenerationFailure:
        logutils.log(f'Density generation failed')
        density_bafdistinfo = None
        density_peakresult = None

    # prepare distinfo and returned peakresult
    if hist:
        bafdistinfo = hist_bafdistinfo 
        peakresult = hist_peakresult
    else:
        if density_bafdistinfo is None:
            bafdistinfo = hist_bafdistinfo 
            peakresult = hist_peakresult
        else:
            bafdistinfo = density_bafdistinfo 
            peakresult = density_peakresult

    # draw plot
    fig, ax = peakutils.draw_peaks(
        bafs, 
        histpeaks=hist_peakresult, 
        densitypeaks=density_peakresult, 
        omit_density=(density_peakresult is None),
        **draw_peak_kwargs,
    )

    if bafdistinfo['homalt'] is not None:
        ax.axvspan(
            bafdistinfo['homalt']['left_ips'], 
            bafdistinfo['homalt']['right_ips'], 
            color='magenta',
            alpha=0.2,
        )
        ax.text(bafdistinfo['homalt']['center'], ax.get_ylim()[1] * 1.05, 'hom')
    if bafdistinfo['hetalt'] is not None:
        ax.axvspan(
            bafdistinfo['hetalt']['left_ips'], 
            bafdistinfo['hetalt']['right_ips'], 
            color='cyan',
            alpha=0.2,
        )
        ax.text(bafdistinfo['hetalt']['center'], ax.get_ylim()[1] * 1.05, 'hom')

    ax.set_xlabel('BAF')
    ax.set_ylabel('density')

    if return_data:
        return fig, ax, bafdistinfo, peakresult
    else:
        return fig, ax


# unused ones

def get_bafdistinfo_base_usetrough(
    bafs, 

    #homalt_cutoff=None, 
    bw_method,
    xs,
    bins,
    reltoheight,

    hist=False,
    return_peakresult=False,
    use_trough=False,

    #homaltpeak_cutoff=0.05,
    **kwargs,
):
    #bafs = np.asarray(bafs)
    #if homalt_cutoff is not None:
    #    bafs = bafs[bafs > homalt_cutoff]

    if hist:
        peakresult = peakutils.find_hist_peaks(
            bafs, 
            reltoheight=reltoheight, 
            bins=bins,
            **kwargs,
        )
    else:
        peakresult = peakutils.find_density_peaks(
            bafs, 
            reltoheight=reltoheight, 
            xs=xs,
            bw_method=bw_method,
            **kwargs,
        )

    # remove contained peaks
    peakresult = peakresult.remove_contained_peaks()

    trough_x = get_bafdistinfo_findtrough(peakresult)
    if (trough_x is not None) and use_trough:
        result = make_result_using_trough(peakresult, reltoheight, trough_x)
    else:
        result = make_result_without_trough(peakresult, reltoheight)
    
    if return_peakresult:
        return result, peakresult
    else:
        return result


def make_result_using_trough(peakresult, reltoheight, trough_x):
    peaks_lt_trough = peakresult.subset(peakresult['peak_xs'] < trough_x)
    peaks_gt_trough = peakresult.subset(peakresult['peak_xs'] > trough_x)
    if (peaks_lt_trough.npeaks == 0) or (peaks_gt_trough.npeaks == 0):
        raise Exception(f'No peaks to the left or right of trough point')

    left_max_peak_idx = peaks_lt_trough.get_highest_peak_index()
    left_max_peak_array_idx = list(peakresult['peak_indexes']).index(left_max_peak_idx)
    left_max_peakinfo = get_bafdistinfo_make_peakinfo(peakresult, left_max_peak_array_idx, reltoheight)

    right_max_peak_idx = peaks_gt_trough.get_highest_peak_index()
    right_max_peak_array_idx = list(peakresult['peak_indexes']).index(right_max_peak_idx)
    right_max_peakinfo = get_bafdistinfo_make_peakinfo(peakresult, right_max_peak_array_idx, reltoheight)

    result = {
        'homalt': left_max_peakinfo,
        'hetalt': right_max_peakinfo,
    }
    return result


def make_result_without_trough(peakresult, reltoheight):
    result = dict()

    if peakresult.npeaks == 1:
        peakinfo = get_bafdistinfo_make_peakinfo(peakresult, 0, reltoheight)
        if check_homalt(peakinfo):
            result['homalt'] = peakinfo
            result['hetalt'] = None
        else:
            result['homalt'] = None
            result['hetalt'] = peakinfo
    else:
        max_peak_indexes = (
            np.argsort(peakresult['peak_properties']['peak_heights'])
        )[::-1][:2]  # (largest_idx, second_largest_idx)
        first_peakinfo = get_bafdistinfo_make_peakinfo(peakresult, max_peak_indexes[0], reltoheight)
        second_peakinfo = get_bafdistinfo_make_peakinfo(peakresult, max_peak_indexes[1], reltoheight)
        first_ishomalt = check_homalt(first_peakinfo)
        second_ishomalt = check_homalt(second_peakinfo)
        if first_ishomalt == second_ishomalt:
            raise FailedNoTroughMethod(
                f'Two maximal peaks are both hom or both het'
            )
        if first_ishomalt:
            result['homalt'] = first_peakinfo
            result['hetalt'] = second_peakinfo
        else:
            result['homalt'] = second_peakinfo
            result['hetalt'] = first_peakinfo

    return result



import pysam

import handygenome.annotation.data as annotdata


class Gene:
    def __init__(self, name, refver):
        self.name = name
        self.refver = refver

        tabixfile = annotdata.TABIXFILES_GENESET[self.refver]
